#!/usr/bin/env python
# -*- coding: utf-8 -*-

from copy import copy
from typing import Dict, Any, Optional

from .base import Block


class Button(Block):
    """
    Render a ``<button>`` with Bootstrap styling.

    Example:

        >>> button = Button(text="My Button")

        When rendered in the template with the ``wildewdigets`` template tag, this
        will produce::

            <button type="button" class="button btn btn-secondary">My Button</button>

    Keyword Args:
        text: The text to use for the button
        color: The Boostrap color class to use for the button
        close: The Boostrap close icon will be used for the button
    """
    block: str = 'button'
    tag: str = 'button'

    #: The Boostrap color class to use for the button
    color: str = 'secondary'
    #: The text to use for the button
    text: str = 'Button'
    #: If ``True``, ignore :py:attr:`text` and make this into a Bootstrap
    #: "close" button with a close icon.
    close: bool = False
    #: If ``True``, ignore :py:attr:`text` and make this into a Bootstrap
    #: "close" button with a close icon.
    size: str = None

    def __init__(
        self,
        text: str = None,
        color: str = None,
        close: bool = None,
        size: str = None,
        **kwargs
    ):
        self.color = color if color else self.color
        self.text = text if text else self.text
        self.close = close if close else self.close
        self.size = size if size else self.size
        if self.close:
            self.text = ""
        super().__init__(self.text, **kwargs)
        self._attributes['type'] = 'button'
        if self.close:
            self.add_class('btn-close')
            self._aria_attributes['label'] = "Close"
        else:
            self.add_class('btn')
            self.add_class(f'btn-{self.color}')
            if self.size:
                self.add_class(f'btn-{self.size}')


class ModalButton(Button):
    """
    Render a ``<button>`` with Bootstrap styling which toggles a Bootstrap modal.

    Example:

        >>> button = ModalButton(text="My Button", target='#mymodal')

        When rendered in the template with the ``wildewdigets`` template tag, this
        will produce::

            <button type="button" class="button button--modal btn btn-secondary" data-toggle="modal"
            data-target="#mymodal">My Button</button>

    Keyword Args:
        target: The CSS target for the Bootstrap modal
    """
    block: str = 'button button--modal'

    #: The CSS target for the Bootstrap modal
    target: Optional[str] = None

    def __init__(
        self,
        target: str = None,
        **kwargs
    ):
        self.target = target if target else self.target
        super().__init__(**kwargs)
        self._data_attributes['toggle'] = 'modal'
        self._data_attributes['target'] = self.target


class CollapseButton(Button):
    """
    Render a ``<button>`` with Bootstrap styling which toggles a ``collapse``.

    Example:

        >>> button = CollapseButton(text="My Button", target='#mymodal')

        When rendered in the template with the ``wildewdigets`` template tag, this
        will produce::

            <button type="button" class="button button--collapse btn btn-secondary"
                data-toggle="collapse" data-target="#mymodal"
                aria-expanded="false aria-controls="mymodal">My Button</button>

    Keyword Args:
        target: The CSS target for the Bootstrap collapse
    """
    block: str = 'button button--collapse'

    #: The CSS target for the Bootstrap collapse
    target: Optional[str] = None

    def __init__(
        self,
        target: str = None,
        **kwargs
    ):
        self.target = target if target else self.target
        super().__init__(**kwargs)
        self._data_attributes['toggle'] = 'collapse'
        self._data_attributes['target'] = self.target
        self._aria_attributes['expanded'] = 'false'
        self._aria_attributes['controls'] = self.target.lstrip("#")


class LinkButton(Button):
    """
    Render an ``<a>`` with Bootstrap button styling which toggles a Bootstrap
    modal.

    Example:

        >>> button = LinkButton(text="My Button", url='https://myexample.com')

        When rendered in the template with the ``wildewdigets`` template tag, this
        will produce::

            <a href="https://myexample.com" class="button button--link btn btn-secondary">My Button</a>


    Keyword Args:
        url: The URL for the ``href`` attribute of the <a> record, defaults to no URL.
    """
    block: str = 'button button--link'
    tag: str = 'a'

    url: Optional[str] = None

    def __init__(self, url: str = None, **kwargs):
        self.url = url if url else self.url
        super().__init__(**kwargs)
        del self._attributes['type']
        if url:
            self._attributes['href'] = self.url


class InputButton(Button):
    """
    Render an ``<input type="submit">`` with Bootstrap button styling.

    Example::

        >>> button = InputButton(text="Save")

    When rendered in the template with the ``wildewdigets`` template tag, this
    will produce::

        <input type="submit" class="button button--submit btn btn-secondary" value="Save">

    """
    template_name: str = 'wildewidgets/block--simple.html'

    block: str = 'button--submit'
    tag: str = 'input'

    confirm_text: str = None

    def __init__(
        self,
        confirm_text: str = None,
        **kwargs
    ):
        self.confirm_text = confirm_text if confirm_text else self.confirm_text
        super().__init__(**kwargs)
        self._attributes['type'] = 'submit'
        self._attributes['value'] = self.text
        if self.confirm_text:
            self._attributes['onclick'] = f"return confirm('{self.confirm_text}');"
        self.text = None


class FormButton(Block):
    """
    Render a ``<form>`` with optional hidden inputs and a submit button.

    Example::

        >>> button = FormButton(
            text="Save",
            action='/my/form/action',
            data={'field1': 'value1'}
        )

    When rendered in the template with the ``wildewdigets`` template tag, this
    will produce::

        <form action="/my/form/action" method="post" class="button-form">
            <input type="hidden" name="csrfmiddlewaretoken" value="__THE_CSRF_TOKEN__">
            <input type="hidden" name="field1" value="value1">
            <input type="submit" class="button button--submit btn btn-secondary" value="Save">
        </form>

    All the constructor parameters can be set in a subclass of this class as class attributes.  Parameters
    to the constructor override any defined class attributes.

    Keyword Args:
        text: The text to use for the button, defaults to 'Button'
        color: The Boostrap color class to use for the button, defaults to 'secondary'
        button_css_class: a string of classes to apply to the button, defaults to no classes.
        button_attributes: Set any additional attributes for the button as key, value pairs,
             defaults to no additional attributes.
        button_data_attributes: Set ``data-`` attributes for the button, defaults to no
            data attributes
    """
    template_name: str = 'wildewidgets/button--form.html'
    tag: str = 'form'
    block: str = 'button-form'

    action: Optional[str] = None
    method: str = 'post'
    text: Optional[str] = None
    color: str = 'secondary'
    button_css_class: Optional[str] = None
    button_attributes: Dict[str, str] = {}
    button_data_attributes: Dict[str, str] = {}
    confirm_text: Optional[str] = None
    data: Dict[str, Any] = {}

    def __init__(self, **kwargs):
        action = kwargs.pop('action', self.action)
        method = kwargs.pop('method', self.method)
        self.close = kwargs.pop('close', False)
        self.data = kwargs.pop('data', copy(self.data))
        self.button = self.__build_button(kwargs)
        super().__init__(**kwargs)
        self._attributes['method'] = method
        self._attributes['action'] = action

    def __build_button(self, kwargs) -> InputButton:
        button_kwargs = {
            'text': kwargs.pop('text', self.text),
            'color': kwargs.pop('color', self.color),
            'css_class': kwargs.pop('button_css_class', self.button_css_class),
            'attributes': kwargs.pop('button_attributes', copy(self.button_attributes)),
            'data_attributes': kwargs.pop('button_data_attributes', copy(self.button_data_attributes))
        }
        if self.close:
            button_kwargs['close'] = True
            button_kwargs['aria_attributes'] = {"label": "Close"}
        button_kwargs['confirm_text'] = kwargs.pop('confirm_text', self.confirm_text)
        return InputButton(**button_kwargs)

    def get_context_data(self, *args, **kwargs) -> Dict[str, Any]:
        context = super().get_context_data(*args, **kwargs)
        context['button'] = self.button
        context['data'] = self.data
        return context


class ButtonRow(Block):
    # FIXME: deprecate this in favor of HorizontalLayoutBlock

    def __init__(self, *blocks, **kwargs):
        css_class = kwargs.get("css_class", "")
        css_class += " d-flex justify-content-end"
        kwargs["css_class"] = css_class
        super().__init__(*blocks, **kwargs)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import warnings
from typing import Optional

try:
    from pygments import highlight
    from pygments.lexers import get_lexer_by_name
    from pygments.formatters import HtmlFormatter
except ModuleNotFoundError:
    # Only needed if using syntax highlighting
    pass

from .base import TemplateWidget, Block


class CodeWidget(Block):
    """
    A widget to display code with syntax highlighting if a language is supplied.

    Keyword Args:
        code: the code to be displayed
        language: the language of the code
        line_numbers: if ``True``, show line numbers
    """
    block: str = 'wildewidgets_highlight_container'

    #: The code to be displayed
    code: str = ""
    #: The programming language for the :py:attr:`code`
    language: Optional[str] = None
    #: If ``True``, show line numbers
    line_numbers: bool = False

    def __init__(
        self,
        code: str = None,
        language: str = None,
        line_numbers: bool = False,
        **kwargs
    ):
        self.code = code if code else self.code
        self.language = language if language else self.language
        if not self.language:
            raise ValueError(
                f'{self.__class__.__name__}: "language" must be defined either as a class attribute or a keyword arg'
            )
        self.line_numbers = line_numbers if line_numbers else self.line_numbers
        super().__init__(**kwargs)
        self.add_code(self.code, language=self.language, line_numbers=self.line_numbers)

    def add_code(self, code: str, language: str, line_numbers: bool = False) -> None:
        lexer = get_lexer_by_name(language)
        formatter = HtmlFormatter(linenos=line_numbers, cssclass="wildewidgets_highlight")
        self.add_block(highlight(code, lexer, formatter))


class MarkdownWidget(TemplateWidget):
    """
    A widget to display markdown as HTML.

    Args:
        text (str): the markdown to render as HTML.
        css_class (str, optional): any classes to add to the widget
    """
    template_name = 'wildewidgets/markdown_widget.html'
    text: str  = ""
    css_class: str = ""

    def __init__(self, *args, **kwargs):
        self.text = kwargs.pop("text", self.text)
        self.css_class = kwargs.pop("css_class", self.css_class)
        super().__init__(*args, **kwargs)

    def get_context_data(self, *args, **kwargs):
        kwargs = super().get_context_data(*args, **kwargs)
        kwargs['text'] = self.text
        kwargs['css_class'] = self.css_class
        return kwargs


class HTMLWidget(TemplateWidget):
    """
    A widget to display raw HTML.

    Args:
        html (str): the HTML to render.
        css_class (str, optional): any classes to add to the widget
    """
    template_name = 'wildewidgets/html_widget.html'
    html = ""
    css_class = None

    def __init__(self, *args, **kwargs):
        self.html = kwargs.pop('html', self.html)
        self.css_class = kwargs.pop("css_class", self.css_class)
        super().__init__(*args, **kwargs)

    def get_context_data(self, *args, **kwargs):
        kwargs = super().get_context_data(*args, **kwargs)
        kwargs['html'] = self.html
        kwargs['css_class'] = self.css_class
        return kwargs


class StringBlock(Block):
    """
    A basic widget that displays a string.

    .. deprecated:: 0.14.0

        Use :py:class:`wildewidgets.widgets.base.Block` directly instead.  It
        works exactly like :py:class:`StringBlock`

    Args:
        text: the text to display.
    """

    def __init__(self, text: str, **kwargs):
        warnings.warn(
            'Deprecated in 0.14.0; use Block directly instead.',
            DeprecationWarning,
            stacklevel=2
        )
        super().__init__(*[text], **kwargs)


class TimeStamp(Block):
    """
    A basic widget that displays a timestamp.

    Args:
        text: the text to display.
    """
    tag = "small"
    css_class = "fw-light"


class TagBlock(Block):
    """
    A basic widget that displays a colored tag.

    Args:
        text: the text to display.
        color: the bootstrap color class.
    """
    block: str = 'badge'

    def __init__(self, text: str, color: str = "secondary", **kwargs):
        super().__init__(text, **kwargs)
        self.add_class(f'bg-{color}')



from copy import deepcopy
import random
from typing import Any, Dict, List, Optional

from django import template
from wildewidgets.views import WidgetInitKwargsMixin

from ..base import Widget

from .components import (
    DataTableColumn,
    DataTableFilter,
    DataTableForm,
    DataTableStyler
)
from .views import DatatableAJAXView


class BaseDataTable(
    Widget,
    WidgetInitKwargsMixin,
    DatatableAJAXView
):
    """
    This is the base class for all other dataTable classes.

    Keyword Args:
        width: The table width as a CSS specifier, e.g. "``100%``", "``500px``"
        height: The table height as a CSS specifier, e.g. "``500px``"
        title: The table title
        searchable: Whether the table is searchable
        paging: Whether the table is paged
        page_length: How many rows should we show on each page?
        small: If ``True``, use smaller font and row height when rendering rows
        buttons: If ``True``, add the dataTable "Copy", "CSV", "Export" and "Print" buttons
        striped: If ``True``, use different colors for even and odd rows
        table_id: The table CSS id
        async: If ``True`` the table will use AJAX to do its work via the ``/wildewidgets_json``.
            If ``False`` all data will be loaded into the table at render time.
        data: The table data
        sort_ascending: If ``True``, sort the rows in the table ascending; if ``False``, sort
            descending.
        action_button_size: The size of the action button. Defaults to 'normal'.
            Valid values are ``normal``, ``sm``, ``lg``.
    """

    template_file: str = 'wildewidgets/table.html'

    #: The URL name for the dataTables AJAX endpoint
    ajax_url_name: str = 'wildewidgets_json'

    # dataTable specific configs

    #: How tall should we make the table?  Any CSS width string is valid.
    height: Optional[str] = None
    #: How wide should we make the table?  Any CSS width string is valid.
    width: Optional[str] = '100%'
    #: If ``True``, use smaller font and row height when rendering rows
    small: bool = False
    #: If ``True``, use different colors for even and odd rows
    striped: bool = False
    #: Show the search input?
    searchable: bool = True
    #: How many rows should we show on each page
    page_length: int = 25
    #: If ``True``, sort rows ascending; otherwise descending.
    sort_ascending: bool = True
    #: Hide our paging, page length and search controls
    hide_controls: bool = False

    #: The CSS id to assign to the table.  The id will be ``datatable_table_{table_id}``
    table_id: Optional[str] = None

    #: If ``True``, add the dataTable "Copy", "CSV", "Export" and "Print" buttons
    buttons: bool = False

    #: Whole table form actions.  If this is not ``None``, add a first column with
    #: checkboxes to each row, and a form that allows you to choose bulk actions
    #: to perform on all checked rows.
    form_actions = None
    #: The URL to which to POST our form actions if :py:attr:`form_actions` is
    #: not ``None``
    form_url: str = ''

    def __init__(
        self,
        *args,
        width: str = None,
        height: str = None,
        title: str = None,
        searchable: str = True,
        paging: str = True,
        page_length: int = None,
        small: bool = None,
        buttons: bool = None,
        striped: bool = None,
        hide_controls: bool = None,
        table_id: str = None,
        sort_ascending: bool = None,
        data: List[Any] = None,
        form_actions: Any = None,
        form_url: str = None,
        ajax_url_name: str = None,
        **kwargs
    ):
        self.width = width if width is not None else self.width
        self.height = height if height is not None else self.height
        self.title = title if title is not None else self.title
        self.searchable = searchable if searchable is not None else self.searchable
        self.page_length = page_length if page_length is not None else self.page_length
        self.small = small if small is not None else self.small
        self.buttons = buttons if buttons is not None else self.buttons
        self.striped = striped if striped is not None else self.striped
        self.hide_controls = hide_controls if hide_controls is not None else self.hide_controls
        #: These are options for dataTable itself and get set in the JavaScript
        #: constructor for the table.
        self.options = {
            'width': self.width,
            'height': self.height,
            "title": self.title,
            "searchable": self.searchable,
            "paging": paging,
            "page_length": self.page_length,
            "small": self.small,
            "buttons": self.buttons,
            "striped": self.striped,
            "hide_controls": self.hide_controls
        }
        #: The CSS id for this table
        self.table_id = table_id if table_id else self.table_id
        if self.table_id is None:
            self.table_id = random.randrange(0, 1000)
        self.table_name = f'datatable_table_{self.table_id}'
        # We have to do this this way instead of naming it above in the kwargs
        # because ``async`` is a reserved keyword
        self.async_if_empty: bool = kwargs.get('async', True)
        #: A mapping of field name to column definition
        self.column_fields: Dict[str, DataTableColumn] = {}
        #: A mapping of field name to column filter definition
        self.column_filters: Dict[str, DataTableFilter] = {}
        #: A list of column styles to apply
        self.column_styles: List[DataTableStyler] = []
        self.data = data if data else []
        self.sort_ascending = sort_ascending if sort_ascending is not None else self.sort_ascending
        self.form_actions = form_actions if form_actions else deepcopy(self.form_actions)
        self.form_url = form_url if form_url else self.form_url
        self.ajax_url_name = ajax_url_name if ajax_url_name else self.ajax_url_name
        if self.has_form_actions():
            self.column_fields['checkbox'] = DataTableColumn(
                field='checkbox',
                verbose_name=' ',
                searchable=False,
                sortable=False
            )
        super().__init__(*args, **kwargs)

    def get_column_number(self, name: str) -> int:
        """
        Return the dataTables column number for the column named ``name``.  You
        might need this if you want to pass it to javascript that will work on
        that column.

        Args:
            name: the name of the column

        Raises:
            IndexError: no column named ``name`` exists in this table

        Returns:
            The 0-offset column number
        """
        columns = list(self.column_fields.keys())
        if name in columns:
            return columns.index(name)
        raise IndexError(f'No column with name "{name}" is registered with this table')

    def has_form_actions(self) -> bool:
        return self.form_actions is not None

    def get_form_actions(self):
        return self.form_actions

    def add_form_action(self, action):
        self.form_actions.append(action)

    def add_column(
        self,
        field: str,
        verbose_name: str = None,
        searchable: bool = True,
        sortable: bool = True,
        align: str = 'left',
        head_align: str = 'left',
        visible: bool = True,
        wrap: bool = True
    ) -> None:
        """
        Add a column to our table.  This updates :py:attr:`column_fields`.

        Args:
            field: the name of the field that to render in this column

        Keyword Args:
            verbose_name: the label to use for the heading of this column
            searchable: if ``True``, include this column in global searches
            sortable: if ``True``, the table can be sorted by this column
            align: horizontal alignment for this column: ``left``, ``right``, ``center``
            visible: if ``False``, the column will be present in the table, but hidden
                from the user
            head_align: horizontal alignment for the header for this column:
                ``left``, ``right``, ``center``
            wrap: if ``True``, wrap contents in this column
        """
        self.column_fields[field] = DataTableColumn(
            field=field,
            verbose_name=verbose_name,
            searchable=searchable,
            sortable=sortable,
            align=align,
            head_align=head_align,
            visible=visible,
            wrap=wrap
        )

    def add_filter(self, field: str, dt_filter: DataTableFilter) -> None:
        """
        Add a column filter.  This updates :py:attr:`column_filters`.

        Args:
            field: the name of the field to filter
            dt_filter: a filter definition
        """
        self.column_filters[field] = dt_filter

    def remove_filter(self, field: str):
        """
        Remove a column filter.  This updates :py:attr:`column_filters`.

        Args:
            field: the name of the field for which to remove the filter
        """
        del self.column_filters[field]

    def add_styler(self, styler: DataTableStyler) -> None:
        styler.test_index = list(self.column_fields.keys()).index(styler.test_cell)
        if styler.target_cell:
            styler.target_index = list(self.column_fields.keys()).index(styler.target_cell)
        self.column_styles.append(styler)

    def build_context(self, **kwargs) -> Dict[str, Any]:
        kwargs['rows'] = self.data
        return kwargs

    def get_template_context_data(self, **kwargs) -> Dict[str, Any]:
        kwargs = super().get_template_context_data(**kwargs)
        has_filters = False
        filters = []
        for key, item in self.column_fields.items():
            if key in self.column_filters:
                filters.append((item, self.column_filters[key]))
                has_filters = True
            else:
                filters.append(None)
        if self.data or not self.async_if_empty:
            kwargs = self.build_context(**kwargs)
            kwargs['async'] = False
        else:
            kwargs['async'] = True
        kwargs['header'] = self.column_fields
        kwargs['has_form_actions'] = self.has_form_actions()
        kwargs['filters'] = filters
        kwargs['stylers'] = self.column_styles
        kwargs['has_filters'] = has_filters
        kwargs['options'] = self.options
        table_id = self.table_id if self.table_id else random.randrange(0, 1000)
        kwargs['name'] = f"datatable_table_{table_id}"
        kwargs['sort_ascending'] = self.sort_ascending
        kwargs["tableclass"] = self.__class__.__name__
        if not self.data:
            kwargs["extra_data"] = self.get_encoded_extra_data()
        kwargs["form"] = DataTableForm(self)
        if 'csrf_token' in kwargs:
            kwargs["csrf_token"] = kwargs['csrf_token']
        kwargs["ajax_url_name"] = self.ajax_url_name
        return kwargs

    def get_content(self, **kwargs) -> str:
        """
        Return the rendered dataTable HTML.

        Keyword Args:
            **kwargs: the template context

        Returns:
            The rendered datatable HTML
        """
        context = self.get_template_context_data(**kwargs)
        html_template = template.loader.get_template(self.template_file)
        return html_template.render(context)

    def __str__(self) -> str:
        return self.get_content()

    def add_row(self, **kwargs) -> None:
        row = []
        for field in self.column_fields:
            if field in kwargs:
                row.append(kwargs[field])
            else:
                row.append('')
        self.data.append(row)

    def render_checkbox_column(self, row: Any, column: str) -> str:
        return f'<input type="checkbox" name="checkbox" value="{row.id}">'



from copy import deepcopy
import datetime
import logging
from typing import Any, Dict, List, Optional, Tuple, Type, Union

from django.conf import settings
from django.core.exceptions import FieldDoesNotExist
from django.db import models
from django.db.models.fields.related import RelatedField, ManyToManyRel

from .base import BaseDataTable
from .actions import ActionButtonBlockMixin, ActionsButtonsBySpecMixin, RowActionButton

logger = logging.getLogger(__name__)


# -------------------------------
# Mixins
# -------------------------------

class ModelTableMixin:
    """
    This mixin is used to create a table from a :py:class:`Model`.  It provides methods
    to create and configure columns based on model fields and configuration.

    Example::

        from django.db import models
        from wildewidgets import DataTable, ModelTableMixin

        class Author(models.Model):

            full_name = models.CharField(...)

        class Book(models.Model):

            title = models.CharField(...)
            isbn = models.CharField(...)
            authors = models.ManyToManyField(Author)

        class BookModelTable(ModelTableMixin, DataTable):

            model = Book

            fields = ['title', 'authors__full_name', 'isbn']
            alignment = {'authors': 'left'}
            verbose_names = {'authors__full_name': 'Authors'}
            buttons = True
            striped = True

            def render_authors__full_name_column(self, row, column):
                authors = row.authors.all()
                if authors.count() > 1:
                    return f"{authors[0].full_name} ... "
                return authors[0].full_name
    """

    #: The Django model class for this table
    model: Optional[Type[models.Model]]

    #: This is either ``None``, the string ``__all__`` or a list of column names
    #: to use in our table.  For the list, entries can either be field names
    #: from our :py:attr:`model`, or names of computed fields that will be
    #: rendered with a ``render_FIELD_column`` method.  If ``None``, empty list
    #: or ``__all__``, display all fields on the :py:attr:`model`.
    fields: Optional[Union[str, List[str]]] = []
    #: The list of field names to hide by default
    hidden: List[str] = []
    #: A mapping of field name to table column heading
    verbose_names: Dict[str, str] = {}
    #: A list of field names that will not be sortable
    unsortable: List[str] = []
    #: A list of field names that will not be searched when doing a global table search
    unsearchable: List[str] = []
    #: A mapping of field name to data type.  This is used to do some automatic formatting
    #: of table values.
    field_types: Dict[str, str] = {}
    #: A mapping of field name to field alignment.  Valid values are ``left``, ``right``, and
    #: ``center``
    alignment: Dict[str, str] = {}
    bool_icons = {}

    def __init__(
        self,
        *args,
        model: Optional[Type[models.Model]] = None,
        fields: Optional[Union[str, List[str]]] = None,
        hidden: List[str] = None,
        verbose_names: List[str] = None,
        unsortable: List[str] = None,
        unsearchable: List[str] = None,
        field_types: Dict[str, str] = None,
        alignment: Dict[str, str] = None,
        bool_icons: Dict[str, str] = None,
        **kwargs
    ):
        self.model = model if model is not None else self.model
        self.fields = fields if fields else deepcopy(self.fields)
        self.hidden = hidden if hidden else deepcopy(self.hidden)
        self.verbose_names = verbose_names if verbose_names else deepcopy(self.verbose_names)
        self.unsortable = unsortable if unsortable else deepcopy(self.unsortable)
        self.unsearchable = unsearchable if unsearchable else deepcopy(self.unsearchable)
        self.field_types = field_types if field_types else deepcopy(self.field_types)
        self.alignment = alignment if alignment else deepcopy(self.alignment)
        self.bool_icons = bool_icons if bool_icons else deepcopy(self.bool_icons)
        super().__init__(*args, **kwargs)

        #: A mapping of field name to Django field class
        self.model_fields: Dict[str, models.Field] = {}
        #: A mapping of field name to Django related field class
        self.related_fields: Dict[str, models.Field] = {}
        #: A list of names of our model fields
        self.field_names: List[str] = []

        # Build our mapping of all known fields on :py:attr:`model`
        for field in self.model._meta.get_fields():
            if field.name == 'id':
                continue
            self.model_fields[field.name] = field
            self.field_names.append(field.name)

        # Find our related fields -- these are in Django QuerySet format, e.g.
        # parent__child or parent__child__grandchild
        for field_name in self.fields:
            if field_name not in self.model_fields:
                field = self.get_related_field(self.model, field_name)
                if field:
                    self.related_fields[field_name] = field

        if not self.fields or self.fields == '__all__':
            self.load_all_fields()
        else:
            for field_name in self.fields:
                self.load_field(field_name)

    def get_related_model(
        self,
        current_model: Type[models.Model],
        field_name: str
    ) -> models.Model:
        return current_model._meta.get_field(field_name).related_model

    def get_related_field(
        self,
        current_model: models.Model,
        name: str
    ) -> Optional[models.Field]:
        """
        Given ``name``, a field specifier in Django QuerySet format, e.g.
        ``parent__child`` or ``parent__child__grandchild``, return the field
        instance that represents that field on the appropriate related model.

        Args:
            current_model:  the model for the leftmost field in ``name``
            name: the field specfier

        Returns:
            the :py:class:`Field` instance, or ``None`` if we couldn't find one.
        """
        name_fields: List[str] = name.split('__')
        if len(name_fields) == 1:
            # This is not a related field specifier
            return None
        # Walk through the fields and find the Django model for the last
        # field in the spec
        for field in name_fields[:-1]:
            current_model = self.get_related_model(current_model, field)
        if current_model:
            try:
                final_field = current_model._meta.get_field(name_fields[-1])
            except FieldDoesNotExist:
                final_field = None
        else:
            final_field = None
        return final_field

    def get_field(self, field_name: str) -> Optional[models.Field]:
        """
        Return the :py:class:`django.db.models.fields.Field` instance for the
        field named ``field_name``.  If no field with that name exists, return
        ``None``.

        Args:
            field_name: the name of the field to retrieve

        Returns:
            A django field instance, or ``None``.
        """
        if field_name in self.model_fields:
            field = self.model_fields[field_name]
        elif field_name in self.related_fields:
            field = self.related_fields[field_name]
        else:
            field = None
        return field

    def set_standard_column_attributes(self, field_name: str, kwargs: Dict[str, Any]) -> None:
        if field_name in self.hidden:
            kwargs['visible'] = False
        if field_name in self.unsearchable:
            kwargs['searchable'] = False
        if field_name in self.unsortable:
            kwargs['sortable'] = False
        if field_name in self.alignment:
            kwargs['align'] = self.alignment[field_name]
        else:
            field = self.get_field(field_name)
            if isinstance(field, (models.TextField, models.CharField, models.DateField, models.DateTimeField)):
                kwargs['align'] = 'left'
            else:
                kwargs['align'] = 'right'

    def load_field(self, field_name: str) -> None:
        if field_name in self.model_fields:
            field = self.model_fields[field_name]
            verbose_name = field.name.replace('_', ' ')
            kwargs = {}
            if field_name in self.verbose_names:
                kwargs['verbose_name'] = self.verbose_names[field_name]
            elif hasattr(field, 'verbose_name'):
                if verbose_name == field.verbose_name:
                    kwargs['verbose_name'] = verbose_name.capitalize()
                else:
                    kwargs['verbose_name'] = field.verbose_name
            self.set_standard_column_attributes(field_name, kwargs)
            self.add_column(field_name, **kwargs)
        else:
            kwargs = {}
            if field_name in self.verbose_names:
                verbose_name = self.verbose_names[field_name]
            else:
                verbose_name = field_name.replace('_', ' ').replace('__', ' ').capitalize()
            kwargs['verbose_name'] = verbose_name
            self.set_standard_column_attributes(field_name, kwargs)
            self.add_column(field_name, **kwargs)

    def load_all_fields(self) -> None:
        for field_name in self.field_names:
            self.load_field(field_name)

    def render_currency_type_column(self, value: Any) -> str:
        return f"${value}"

    def render_bool_type_column(self, value: Any) -> str:
        if value == "True":
            return '<i class="bi-check-lg text-success"><span style="display:none">True</span></i>'
        return ""

    def render_bool_icon_column(self, value: Any, icon_data: Tuple[str, str]) -> str:
        if len(icon_data) == 0:
            return value
        if value == "True":
            return f"<i class='bi-{icon_data[0][0]} {icon_data[0][1]}'><span style='display:none'>True</span></i>"
        if len(icon_data) > 1:
            return f"<i class='bi-{icon_data[1][0]} {icon_data[1][1]}'><span style='display:none'>False</span></i>"
        return ""

    def render_datetime_type_column(self, value: datetime.datetime) -> str:
        datetime_format = "%m/%d/%Y %H:%M"
        if hasattr(settings, 'WILDEWIDGETS_DATETIME_FORMAT'):
            datetime_format = settings.WILDEWIDGETS_DATETIME_FORMAT
        if value:
            return value.strftime(datetime_format)
        return ""

    def render_date_type_column(self, value: datetime.date) -> str:
        date_format = "%m/%d/%Y"
        if hasattr(settings, 'WILDEWIDGETS_DATE_FORMAT'):
            date_format = settings.WILDEWIDGETS_DATE_FORMAT
        if value:
            return value.strftime(date_format)
        return ""

    def render_column(self, row: Any, column: str) -> str:
        value = super().render_column(row, column)
        if column in self.model_fields:
            field = self.model_fields[column]
        elif column in self.related_fields:
            field = self.related_fields[column]
        else:
            field = None
        if column in self.field_types:
            field_type = self.field_types[column]
            attr_name = f"render_{field_type}_type_column"
            if hasattr(self, attr_name):
                if attr_name in ['date', 'datetime']:
                    value = getattr(row, column)
                return getattr(self, attr_name)(value)
        elif isinstance(field, models.DateTimeField):
            return self.render_datetime_type_column(getattr(row, column))
        elif isinstance(field, models.DateField):
            return self.render_date_type_column(getattr(row, column))
        elif column in self.bool_icons:
            return self.render_bool_icon_column(value, self.bool_icons[column])
        return value


class WidgetCellMixin:
    """
    This mixin is used to display a widget in a cell based on the value of the
    column.
    """

    # Specify the widget to use for each column. The key is the column name
    # and the value is the widget class. The widget class must be a subclass
    # of :py:class:`wildewidgets.widgets.base.Block`. Fields that are not specified
    # will not have a widget, but will be displayed as text.
    cell_widgets: Dict[str, str] = {}

    def render_column(self, row: Any, column: str) -> str:
        if column in self.cell_widgets:
            widget_class = self.cell_widgets[column]
            widget = widget_class(row=row, column=column)
            return str(widget)
        return super().render_column(row, column)


# -------------------------------
# Tables
# -------------------------------

class DataTable(
    ActionsButtonsBySpecMixin,
    BaseDataTable
):
    pass


class BasicModelTable(ModelTableMixin, DataTable):
    """
    This class is used to create a table from a
    :py:class:`django.db.models.Model`.  It provides a full featured table with
    a minimum of code. Many derived classes will only need to define class
    variables.

    Example::

        class BookModelTable(BasicModelTable):
            fields = ['title', 'authors__full_name', 'isbn']
            model = Book
            alignment = {'authors': 'left'}
            verbose_names = {'authors__full_name': 'Authors'}
            buttons = True
            striped = True

            def render_authors__full_name_column(self, row, column):
                authors = row.authors.all()
                if authors.count() > 1:
                    return f"{authors[0].full_name} ... "
                return authors[0].full_name
    """
    pass


class ActionButtonModelTable(
    ActionButtonBlockMixin,
    ModelTableMixin,
    BaseDataTable
):
    """
    This class is used to create a table from a
    :py:class:`django.db.models.Model`, and allows you to specify per-row actions
    as :py:class:`wildewidgets.widgets.tables.actions.RowActionButton` classes or
    subclasses.

    Example::

        class BookModelTable(ActionButtonModelTable):
            fields = ['title', 'authors__full_name', 'isbn']
            model = Book
            alignment = {'authors': 'left'}
            verbose_names = {'authors__full_name': 'Authors'}
            striped = True

            actions = [
                RowModelUrlButton(text='Edit', color='primary', attribute='get_absolute_url')
            ]

            def render_authors__full_name_column(self, row, column):
                authors = row.authors.all()
                if authors.count() > 1:
                    return f"{authors[0].full_name} ... "
                return authors[0].full_name
    """
    pass


class StandardActionButtonModelTable(
    ActionButtonBlockMixin,
    ModelTableMixin,
    BaseDataTable
):
    """
    This class is used to create a table from a
    :py:class:`django.db.models.Model`, and allows you to specify per-row
    actions as :py:class:`wildewidgets.widgets.tables.actions.RowActionButton`
    classes or subclasses.  An "Edit" and "Delete" button will be created for
    each row automatically.

    Important:

        For the buttons to work, use the
        :py:class:`wildewidgets.models.ViewSetMixin` on your model, and
        define the ``get_update_url`` and ``get_delete_url`` methods.

    Example::

        from django.db import models
        from django.urls import reverse
        from wildewdigets.models import ViewSetMixin
        from wildewidgets import StandardActionButtonModelTable

        class Book(ViewSetMixin, models.Model):

            title = models.CharField('Title', ...)
            isbn = models.CharField('ISBN', ...)
            publisher = models.CharField('Publisher', ...)
            num_pages = models.IntegerField('# Pages', ...)

            def get_update_url(self):
                return reverse('core:book--edit', args=[self.id])

            def get_delete_url(self):
                return reverse('core:book--delete', args=[self.id])


        class BookModelTable(StandardActionButtonModelTable):
            fields = ['title', 'publisher', 'isbn', 'num_pages']
            model = Book
            striped = True
    """
    pass


class LookupModelTable(
    ActionButtonBlockMixin,
    ModelTableMixin,
    BaseDataTable
):
    """
    This class is used to create a table for a lookup table.  It from
    :py:class:`ActionButtonsModelTable` in that if :py:attr:`fields` is empty,
    we'll build our ``fields`` list such that:

    * Only non-related fields are shown in the table
    * the ``id`` column is left aligned

    """

    actions: List[RowActionButton] = []

    def __init__(self, *args, **kwargs):
        fields = kwargs.pop('fields', [])
        model = kwargs.get('model')
        if not model:
            model = self.model
        if not fields and not self.fields:
            alignment = {}
            fields = []
            for field in model._meta.get_fields():
                if not isinstance(field, (RelatedField, ManyToManyRel)):
                    # Exclude any RelatedField -- that's what
                    fields.append(field.name)
                    if field.name == 'id':
                        alignment['id'] = 'left'
            kwargs['fields'] = fields
            kwargs['alignment'] = alignment
        super().__init__(*args, **kwargs)



from copy import copy, deepcopy

from typing import Any, Dict, List, Optional, TYPE_CHECKING

from django.urls import reverse

from ..base import Block
from ..forms import HiddenInputBlock
from ..buttons import InputButton
from ..structure import HorizontalLayoutBlock

if TYPE_CHECKING:
    from django.contrib.auth.models import AbstractUser


# -------------------------------
# Buttons
# -------------------------------

class RowActionButton(Block):

    tag: str = 'a'

    text: Optional[str] = None
    url: Optional[str] = None
    color: str = 'secondary'
    size: str = None
    permission: Optional[str] = None

    def __init__(
        self,
        text: str = None,
        url: str = None,
        color: str = None,
        size: str = None,
        permission: str = None,
        **kwargs
    ):
        self.text = text if text else self.text
        self.url = url if url else self.url
        self.color = color if color else self.color
        self.size = size if size else self.size
        self.permission = permission if permission else self.permission
        super().__init__(self.text, **kwargs)

        #: The table will set this
        self.row: Any = None

    def is_visible(self, row: Any, user: "Optional[AbstractUser]") -> bool:  # pylint: disable=arguments-differ
        if self.permission is not None and user is not None:
            if user.has_perm(self.permission):
                return True
            return False
        return True

    def get_url(self, row: Any) -> str:
        return self.url

    def bind(
        self,
        row: Any,
        table: "ActionButtonBlockMixin",
        size: str = None
    ) -> "RowActionButton":
        """
        Bind this button to a particular ``row``, and return the bound button.

        Note:
            We take ``table`` as an argument so that we can get access to attributes
            that are only on the containing table, e.g. ``csrf_token``, which is set
            during the AJAX call to our Table.

        Args:
            row: the row we're working on
            table: the DataTable instance that owns the row

        Keyword Args
            size: set the Bootstrap size of the button to this

        Returns:
            A copy of this button, bound to ``row``.
        """
        action = deepcopy(self)
        action.row = row
        action.add_class(f'btn-{self.color}')
        action._attributes['href'] = self.get_url(row)
        if not size:
            size = self.size
        if self.size:
            action.add_class(f'btn-{self.size}')
        return action


class RowLinkButton(RowActionButton):

    def get_context_data(self, *args, **kwargs) -> Dict[str, Any]:
        self._attributes['href'] = self.get_url(self.row)
        self._attributes['role'] = 'button'
        return super().get_context_data(*args, **kwargs)


class RowModelUrlButton(RowActionButton):

    block: str = 'btn'
    modifier: str = 'auto'

    attribute: str = 'get_absolute_url'
    text: Optional[str] = 'View'

    def __init__(
        self,
        attribute: str = None,
        **kwargs
    ):
        self.attribute = attribute if attribute else self.attribute
        super().__init__(**kwargs)

    def get_url(self, row: Any) -> str:
        if not hasattr(row, self.attribute):
            raise ValueError(
                f'{row} has no "{self.attribute}" attribute or method'
            )
        attr = getattr(row, self.attribute)
        if callable(attr):
            return attr()
        return attr


class RowDjangoUrlButton(RowActionButton):

    block: str = 'btn'

    url_path: Optional[str] = None
    url_args: List[str] = []
    url_kwargs: Dict[str, str] = {}

    def __init__(
        self,
        url_path: str = None,
        url_args: List[str] = None,
        url_kwargs: Dict[str, str] = None,
        **kwargs
    ):
        self.url_path = url_path if url_path else self.url_path
        self.url_args = url_args if url_args else copy(self.text)
        self.url_kwargs = url_kwargs if url_kwargs else copy(self.url_kwargs)
        if not self.url_path:
            raise self.RequiredAttrOrKwarg('url_path')
        if self.url_args and self.url_kwargs:
            raise ValueError(
                'Either "url_args" or "url_kwargs" must be provided as keyword arguments or '
                'class attributes, but not both'
            )
        super().__init__(**kwargs)

    def get_url(self, row: Any) -> str:
        if not (self.args and self.kwargs):
            url = reverse(self.url_path)
        if self.args:
            args = [getattr(row, arg) for arg in self.args]
            url = reverse(self.url_path, args=args)
        if self.kwargs:
            kwargs = {kwarg: getattr(self.row, attr) for kwarg, attr in self.kwargs}
            url = reverse(self.url_path, kwargs=kwargs)
        return url


class RowFormButton(RowModelUrlButton):

    block: str = 'action-form'
    tag: str = 'form'

    form_fields: List[str] = []
    confirm_text: str = None

    def __init__(
        self,
        form_fields: List[str] = None,
        confirm_text: str = None,
        **kwargs
    ):
        self.form_fields = form_fields if form_fields else deepcopy(self.form_fields)
        self.confirm_text = confirm_text if confirm_text else self.confirm_text
        super().__init__(**kwargs)
        # Our superclass added self.text as a block; but that will not work for us,
        # since we want self.text to be the button text, so remove it.
        self.blocks = []
        self._attributes['method'] = 'post'

    def get_confirm_text(self, row: Any) -> str:
        return self.confirm_text

    def bind(self, row: Any, table: Any, size: str = None) -> "RowFormButton":
        if not size:
            size = self.size
        action = deepcopy(self)
        action.add_block(HiddenInputBlock(input_name='csrfmiddlewaretoken', value=table.csrf_token))
        action._attributes['action'] = self.get_url(row)
        for field in self.form_fields:
            action.add_block(HiddenInputBlock(input_name=field, value=getattr(row, field)))
        action.add_block(
            InputButton(
                text=self.text,
                color=self.color,
                size=size,
                confirm_text=self.get_confirm_text(row)
            )
        )
        return action


class RowEditButton(RowModelUrlButton):

    attribute: str = 'get_update_url'
    text: str = 'Edit'
    color: str = 'primary'


class RowDeleteButton(RowFormButton):

    attribute: str = 'get_delete_url'
    text: str = 'Delete'
    color: str = 'outline-secondary'
    form_fields: List[str] = ['id']

    def get_confirm_text(self, row: Any) -> str:
        return f'Delete "{str(row)}"?'


# -------------------------------
# Mixins
# -------------------------------

class ActionsButtonsBySpecMixin:
    """
    This is a mixin class for :py:class:`DataTable` classes that allows you to specify
    buttons that will appear per row in an "Actions" column, rightmost of all columns.

    If :py:attr:`actions` is ``True``, and a row object has an attribute or
    method named ``get_absolute_url``,  render a single button named
    :py:attr:`default_action_button_label` that when clicked takes the user
    to that URL.

    Otherwise, :py:attr:`actions` should be a iterable of tuples or lists of
    the following structures::

        (label, Django URL name)
        (label, Django URL name, 'get' or 'post')
        (label, Django URL name, 'get' or 'post', color)
        (label, Django URL name, 'get' or 'post', color)
        (label, Django URL name, 'get' or 'post', color, pk_field_name)
        (label, Django URL name, 'get' or 'post', color, pk_field_name, javascript function name)

    Note:
        The magic is all done in :py:meth:`render_actions_column`, so if you want
        to see what's really going on, read that code.

    Examples:

        To make a button named "Edit" that goes to the django URL named
        ``core:model--edit``::

            ('Edit', 'core:model--edit')

        This renders a "button" something like this::

            <a href="/core/model/edit/?id=1" class="btn btn-secondary me-2">Edit</a>

        Make a button named "Delete" that goes to the django URL named
        ``core:model--delete`` as a ``POST``::

            ('Delete', 'core:model--delete', 'post')

        This renders a "button" something like this::

            <form class"form form-inline" action="/core/model/delete/" method="post">
                <input type="hidden" name="csrfmiddlewaretoken" value="__THE_TOKEN__">
                <input type="hidden" name="id" value="1">
                <input type="submit" value="Delete" class="btn btn-secondary me-2">
            </form>
    """

    #: Per row action buttons.  If not ``False``, this will simply add a
    #: rightmost column  named ``Actions`` with a button named
    #: :py:attr:`default_action_button_label` which when clicked will take the
    #: user to the
    actions: Any = False
    #: How big should each action button be? One of ``normal``, ``btn-lg``, or ``btn-sm``.
    action_button_size: str = 'normal'
    #: The label to use for the default action button
    default_action_button_label: str = 'View'
    #: The Bootstrap color class to use for the default action buttons
    default_action_button_color_class: str = 'secondary'

    def __init__(
        self,
        *args,
        actions: Any = None,
        action_button_size: str = None,
        default_action_button_label: str = None,
        default_action_button_color_class: str = None,
        **kwargs
    ):
        self.actions = actions if actions is not None else self.actions
        self.action_button_size = action_button_size if action_button_size else self.action_button_size
        self.default_action_button_label = (
            default_action_button_label
            if default_action_button_label else self.default_action_button_label
        )
        self.default_action_button_color_class = (
            default_action_button_color_class
            if default_action_button_color_class else self.default_action_button_color_class
        )
        if self.action_button_size != 'normal':
            self.action_button_size_class = f"btn-{self.action_button_size}"
        else:
            self.action_button_size_class = ''
        super().__init__(*args, **kwargs)

    def get_template_context_data(self, **kwargs) -> Dict[str, Any]:
        if self.actions:
            self.add_column(field='actions', searchable=False, sortable=False)
            kwargs['has_actions'] = True
            kwargs['action_column'] = len(self.column_fields) - 1
        else:
            kwargs['has_actions'] = False
        return super().get_template_context_data(**kwargs)

    def get_content(self, **kwargs) -> str:
        if self.actions:
            self.add_column(field='actions', searchable=False, sortable=False)
        return super().get_content(**kwargs)

    def get_action_button(
        self,
        row: Any,
        label: str,
        url_name: str,
        method: str = 'get',
        color_class: str = 'secondary',
        attr: str = 'id',
        js_function_name: str = None
    ) -> str:
        if url_name:
            base = reverse(url_name)
            # FIXME: This assumes we're using QueryStringKwargsMixin, which people
            # outside our group don't use
            if method == 'get':
                url = f"{base}?{attr}={row.id}"
            else:
                url = base
        else:
            url = "javascript:void(0)"
        return self.get_action_button_with_url(row, label, url, method, color_class, attr, js_function_name)

    def get_action_button_url_extra_attributes(self, row: Any) -> str:
        return ""

    def get_action_button_with_url(
        self,
        row: Any,
        label: str,
        url: str,
        method: str = 'get',
        color_class: str = 'secondary',
        attr: str = 'id',
        js_function_name: str = None
    ) -> str:
        url_extra = self.get_action_button_url_extra_attributes(row)
        if url_extra:
            url = f"{url}&{url_extra}"
        if method == 'get':
            if js_function_name:
                link_extra = f'onclick="{js_function_name}({row.id});"'
            else:
                link_extra = ""
            return f'<a href="{url}" class="btn btn-{color_class} {self.action_button_size_class} me-2" {link_extra}>{label}</a>'
        token_input = f'<input type="hidden" name="csrfmiddlewaretoken" value="{self.csrf_token}">'
        id_input = f'<input type="hidden" name="{attr}" value="{row.id}">'
        button = f'<input type=submit value="{label}" class="btn btn-{color_class} {self.action_button_size_class} me-2">'
        form = f'<form class="form form-inline" action={url} method="post">{token_input}{id_input}{button}</form>'
        return form

    def get_conditional_action_buttons(self, row: Any) -> str:
        return ''

    def render_actions_column(self, row: Any, column: str) -> str:
        """
        Render the buttons in the "Actions" column.  This will only be called if
        :py:attr:`actions` is not falsy.  We rely on :py:attr:`actions` to be
        specified in a particular way in order for this to work; see
        :py:class:`ActionButtonsBySpecMixin` for information about how to
        specify button specs in :py:attr:`actions` is constructed.

        Args:
            row: the row data for which we are building action buttons
            column: unused

        Returns:
            The HTML to render into the "Actions" column for ``row``.
        """
        response = '<div class="d-flex flex-row justify-content-end">'
        if hasattr(row, 'get_absolute_url'):
            if callable(row.get_absolute_url):
                url = row.get_absolute_url()
            else:
                url = row.get_absolute_url
            view_button = self.get_action_button_with_url(
                row,
                self.default_action_button_label,
                url,
                color_class=self.default_action_button_color_class
            )
            response += view_button
        if not isinstance(self.actions, bool):
            for action in self.actions:
                if not len(action) > 1:
                    continue
                label = action[0]
                url_name = action[1]
                if len(action) > 2:
                    method = action[2]
                else:
                    method = 'get'
                if len(action) > 3:
                    color_class = action[3]
                else:
                    color_class = 'secondary'
                if len(action) > 4:
                    attr = action[4]
                else:
                    attr = 'id'
                if len(action) > 5:
                    js_function_name = action[5]
                else:
                    js_function_name = ''
                response += self.get_action_button(row, label, url_name, method, color_class, attr, js_function_name)
        response += self.get_conditional_action_buttons(row)
        response += "</div>"
        return response


class ActionButtonBlockMixin:
    """
    This is a mixin class for dataTable classes that allows you to specify
    buttons that will appear per row in an "Actions" column, rightmost of all
    columns.

    """

    #: If not ``None``, make all per-row action buttons be this size.
    button_size: Optional[str] = None
    justify: str = 'end'

    actions: List[RowActionButton] = [
        RowModelUrlButton(text='View', color='secondary')
    ]

    def __init__(
        self,
        *args,
        actions: List[RowActionButton] = None,
        button_size: str = None,
        justify: str = None,
        **kwargs
    ):
        self.actions = actions if actions is not None else deepcopy(self.actions)
        self.button_size = button_size if button_size else self.button_size
        self.justify = justify if justify else self.justify
        super().__init__(*args, **kwargs)

    def get_actions(self) -> List[RowActionButton]:
        return self.actions

    def get_content(self, **kwargs) -> str:
        if self.get_actions():
            self.add_column(field='actions', searchable=False, sortable=False)
        return super().get_content(**kwargs)

    def render_actions_column(self, row: Any, column: str) -> str:
        container = HorizontalLayoutBlock(justify=self.justify)
        for action in self.get_actions():
            button = action.bind(row, self, size=self.button_size)
            user = None
            if hasattr(self, 'request') and self.request is not None:
                user = self.request.user
            if not button.is_visible(row, user):
                continue
            button.add_class('me-2')
            container.add_block(button)
        button.remove_class('me-2')
        return str(container)


class StandardModelActionButtonBlockMixin(ActionButtonBlockMixin):
    """
    This is an :py:class:`ActionButtonBlockMixin` that supplies an "Edit" and
    "Delete" button for each row.

    Important:

        For this to work, use the
        :py:class:`wildewidgets.models.ViewSetMixin` on your model, and
        define the ``get_update_url`` and ``get_delete_url`` methods.
    """

    actions: List[RowActionButton] = [
        RowModelUrlButton(text='Edit', color='primary', attribute='get_update_url'),
        RowModelUrlButton(text='delete', color='outline-secondary', attribute='get_delete_url'),
    ]



import importlib
import os

from django.apps import apps
from django.http import JsonResponse
from django.views.generic import View
from django.views.generic.base import TemplateView

from .mixins import WidgetInitKwargsMixin, JSONResponseMixin


class JSONResponseView(JSONResponseMixin, TemplateView):
    pass


class JSONDataView(View):

    def get(self, request, *args, **kwargs):
        context = self.get_context_data()
        return self.render_to_response(context)

    def get_context_data(self, **kwargs):
        return {}

    def render_to_response(self, context, **response_kwargs):
        return JsonResponse(context)


class WildewidgetDispatch(WidgetInitKwargsMixin, View):

    def dispatch(self, request, *args, **kwargs):
        wildewidgetclass = request.GET.get('wildewidgetclass', None)
        csrf_token = request.GET.get('csrf_token', '')
        if wildewidgetclass:
            configs = apps.get_app_configs()
            for config in configs:
                check_file = os.path.join(config.path, "wildewidgets.py")
                check_dir = os.path.join(config.path, "wildewidgets")
                if os.path.isfile(check_file) or os.path.isdir(check_dir):
                    module = importlib.import_module(f"{config.name}.wildewidgets")
                    if hasattr(module, wildewidgetclass):
                        class_ = getattr(module, wildewidgetclass)
                        extra_data = self.get_decoded_extra_data(request)
                        initargs = extra_data.get('args', [])
                        initkwargs = extra_data.get('kwargs', {})
                        instance = class_(*initargs, **initkwargs)
                        instance.request = request
                        instance.csrf_token = csrf_token
                        instance.args = initargs
                        instance.kwargs = initkwargs
                        return instance.dispatch(request, *args, **kwargs)



from typing import Any, Dict, Optional, TYPE_CHECKING
import base64
import json

from django.core.serializers.json import DjangoJSONEncoder
from django.http import HttpResponse
from django.utils.cache import add_never_cache_headers
try:
    from django.utils.encoding import force_text
except ImportError:
    from django.utils.encoding import force_str as force_text  # type: ignore
from django.utils.functional import Promise

if TYPE_CHECKING:
    from ..widgets import BreadcrumbBlock


class LazyEncoder(DjangoJSONEncoder):
    """
    Encodes django's lazy i18n strings.
    """
    def default(self, o):
        if isinstance(o, Promise):
            return force_text(o)
        return super(LazyEncoder, self).default(o)


class WidgetInitKwargsMixin:

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.extra_data = {
            "args": args,
            "kwargs": kwargs
        }

    def get_encoded_extra_data(self):
        data_json = json.dumps(self.extra_data)
        payload_bytes = base64.b64encode(data_json.encode())
        payload = payload_bytes.decode()
        return payload

    def get_decoded_extra_data(self, request):
        encoded_extra_data = request.GET.get("extra_data", None)
        if not encoded_extra_data:
            return {}
        extra_bytes = encoded_extra_data.encode()
        payload_bytes = base64.b64decode(extra_bytes)
        payload = json.loads(payload_bytes.decode())
        return payload

    def convert_extra(self, extra_item, first=True):
        if first:
            start = '?'
        else:
            start = '&'
        if isinstance(extra_item, dict):
            extra_list = []
            for k, v in extra_item.items():
                extra_list.append(f"{k}={v}")
            extra = f"{start}{'&'.join(extra_list)}"
            return extra
        return ''


class JSONResponseMixin:

    is_clean: bool = False

    def render_to_response(self, context: Dict[str, Any]) -> str:
        """
        Returns a JSON response containing 'context' as payload
        """
        return self.get_json_response(context)

    def get_json_response(self, content: Any, **httpresponse_kwargs) -> HttpResponse:
        """
        Construct an `HttpResponse` object.
        """
        response = HttpResponse(
            content,
            content_type='application/json',
            **httpresponse_kwargs
        )
        add_never_cache_headers(response)
        return response

    def post(self, *args, **kwargs):
        return self.get(*args, **kwargs)

    def get(self, request, *args, **kwargs):
        self.request = request
        self.csrf_token = self.request.GET.get('csrf_token', None)
        response = None

        func_val = self.get_context_data(**kwargs)
        if not self.is_clean:
            assert isinstance(func_val, dict)
            response = dict(func_val)
            if 'error' not in response and 'sError' not in response:
                response['result'] = 'ok'
            else:
                response['result'] = 'error'
        else:
            response = func_val
        # can't have 'view' here, because the view object can't be jsonified
        response.pop('view', None)

        dump = json.dumps(response, cls=LazyEncoder)
        return self.render_to_response(dump)


class StandardWidgetMixin:
    """
    A class based view mixin for views that use a standard widget template.
    This is used with a template-less design.

    The template used by your derived class should include at least the following::

        {% extends "<your_base_template>.html" %}
        {% load  wildewidgets %}

        {% block title %}{{page_title}}{% endblock %}

        {% block breadcrumb-items %}
        {% if breadcrumbs %}
            {% wildewidgets breadcrumbs %}
        {% endif %}
        {% endblock %}

        {% block content %}
        {% wildewidgets content %}
        {% endblock %}

    The ``content`` block is where the content of the page is rendered. The
    ``breadcrumbs`` block is where the breadcrumbs are rendered.

    An example derived class, which will be used by most of the views in the project::

        class DemoStandardMixin(StandardWidgetMixin, NavbarMixin):
            template_name='core/standard.html'
            menu_class = DemoMenu

    The ``DemoStandardMixin`` class is used in the following way::

        class HomeView(DemoStandardMixin, TemplateView):
            menu_item = 'Home'

            def get_content(self):
                return HomeBlock()

            def get_breadcrumbs(self):
                breadcrumbs = DemoBaseBreadcrumbs()
                breadcrumbs.add_breadcrumb('Home')
                return breadcrumbs
    """

    def get_context_data(self, **kwargs):
        kwargs['content'] = self.get_content()
        breadcrumbs: "Optional[BreadcrumbBlock]" = self.get_breadcrumbs()
        if breadcrumbs:
            kwargs['breadcrumbs'] = breadcrumbs
            kwargs['page_title'] = breadcrumbs.flatten()
        return super().get_context_data(**kwargs)

    def get_content(self):
        raise NotImplementedError(
            f"You must override get_content in {self.__class__.__name__}"
        )

    def get_breadcrumbs(self) -> "Optional[BreadcrumbBlock]":
        return None



from typing import Optional

from .base import Block


class FontIcon(Block):
    """
    Render a font-based Bootstrap icon, for example::

        <i class="bi-star"></i>

    See the `Boostrap Icons <https://icons.getbootstrap.com/>`_ list for the
    list of icons.  Find an icon you like, and use the name of that icon on that
    page as the ``icon`` kwarg to the constructor, or set it as the :py:attr:`icon`
    class variable.

    Keyword Args:
        icon: the name of the icon to render, from the Bootstrap Icons list
        color: use this as Tabler color name to use as the foreground
            font color, leaving the background transparent.  If ``background``
            is also set, this is ignored.  Look at `Tabler: Colors
            <https://preview.tabler.io/docs/colors.html>`_
            for your choices; set this to the text after the ``bg-``
        background: use this as Tabler background/foreground color set for
            this icon.  : This overrides :py:attr:`color`. Look
            at `Tabler: Colors <https://preview.tabler.io/docs/colors.html>`_
            for your choices; set this to the text after the ``bg-``
    """

    tag: str = 'i'
    block: str = 'fonticon'

    #: The icon font family prefix.  One could override this to use FontAwesome icons,
    #: for instance, buy changing it to ``fa``
    prefix: str = 'bi'

    #: Use this as the name for the icon to render
    icon: Optional[str] = None
    #: If not ``None``, use this as Tabler color name to use as the foreground
    #: font color, leaving the background transparent.  If :py:attr:`background`
    #: is also set, this is ignored.  Look at `Tabler: Colors
    # <https://preview.tabler.io/docs/colors.html>`_  for your choices; set
    #: this to the text after the ``bg-``
    color: Optional[str] = None
    #: If not ``None``, use this as Tabler background/foreground color set for
    #: this icon.  : This overrides :py:attr:`color`. Look
    #: at `Tabler: Colors <https://preview.tabler.io/docs/colors.html>`_
    #: for your choices; set this to the text after the ``bg-``
    background: Optional[str] = None

    def __init__(
        self,
        icon: Optional[str] = None,
        color: Optional[str] = None,
        background: Optional[str] = None,
        **kwargs
    ) -> None:
        super().__init__(**kwargs)
        self.color = color if color else self.color
        self.background = background if background else self.background
        self.icon = icon if icon else self.icon
        if not self.icon:
            raise ValueError('"icon" is required as either a keyword argument or as a class attribute')
        self.icon = f'{self.prefix}-{icon}'
        self.add_class(self.icon)
        if self.color:
            self.add_class(f'text-{self.color} bg-transparent')
        elif self.background:
            self.add_class(f' bg-{self.background} text-{self.background}-fg')


class TablerFontIcon(FontIcon):
    """
    :py:class:`FontIcon` for Tabler Icons.

    Requires::

        <link rel="stylesheet" href="https://cdn.jsdelivr.net/npm/@tabler/icons@latest/iconfont/tabler-icons.min.css">
    """
    prefix = 'ti ti'


class TablerMenuIcon(FontIcon):
    """
    A Tabler menu specific icon.  This just adds some menu specific classes and
    uses a ``<span>`` instead of a ``<i>``.  It is used by
    :py:class:`wildewidgets.NavItem`, :py:class:`wildewidgets.NavDropdownItem`
    and :py:class:`wildewidgets.DropdownItem` objects.

    Typically, you won't use this directly, but instead it will be created for
    you from a :py:class:`wildewdigets.MenuItem` specification when
    :py:class:`wildewdigets.MenuItem.icon` is not ``None``.

    Example:

        >>> icon = TablerMenuIcon(icon='target')
        >>> item = NavItem(text='Page', url='/page', icon=icon)
        >>> item2 = DropdownItem(text='Page', url='/page', icon=icon)
        >>> item3 = NavDropdownItem(text='Page', url='/page', icon=icon)
    """

    tag: str = 'span'
    block: str = 'nav-link-icon'
    css_class: str = 'd-md-none d-lg-inline-block'



from functools import lru_cache
import logging
import re

from typing import Any, Dict, List, Optional, Tuple, Type, Union

from django.db import models
from django.db.models import Q
from django.utils.html import escape

from wildewidgets.views import JSONResponseView

logger = logging.getLogger(__name__)


class DatatableMixin:

    #: Use this
    model: Optional[Type[models.Model]] = None

    #: The names of the columns to display in our table
    columns: List[str] = []
    #: internal cache for columns definition
    _columns = []
    #: The list of column names by which the table will be ordered.
    order_columns: List[str] = []

    #: The max number of of records that will be returned, so that we can protect our
    #: our server from rendering huge amounts of data
    max_display_length: int = 100
    #: Set to ``True`` if you are using datatables < 1.10
    pre_camel_case_notation: bool = False
    #: Replace any column value that is ``None`` with this string
    none_string: str = ''
    # If set to ``True`` then values returned by :py:meth:`render_column`` will be escaped
    escape_values: bool = True
    #: This gets set by the dataTables Javascript when it does the AJAX call
    columns_data: List[Dict[str, Any]] = []
    #: This determines the type of results. If the AJAX call passes us a data attribute
    #: that is not an integer, it expects a dictionary with specific fields in
    #: the response, see: `dataTables columns.data
    #: <https://datatables.net/reference/option/columns.data>`_
    is_data_list: bool = True

    FILTER_ISTARTSWITH: str = 'istartswith'
    FILTER_ICONTAINS: str = 'icontains'

    def __init__(
        self,
        model: Type[models.Model] = None,
        order_columns: List[str] = None,
        max_display_length: int = None,
        none_string: str = None,
        is_data_list: bool = None,
        escape_values: bool = None,
        **kwargs
    ):
        self.model = model if model else self.model
        self.order_columns = order_columns if order_columns else self.order_columns
        self.max_display_length = max_display_length if max_display_length else self.max_display_length
        self.none_string = none_string if none_string else self.none_string
        self.is_data_list = is_data_list if is_data_list else self.is_data_list
        self.escape_values = escape_values if escape_values else self.escape_values
        super().__init__(**kwargs)

    @property
    def _querydict(self):
        if self.request.method == 'POST':
            return self.request.POST
        return self.request.GET

    def get_initial_queryset(self) -> models.QuerySet:
        if not self.model:
            raise NotImplementedError(
                "Need to provide a model or implement get_initial_queryset!"
            )
        return self.model.objects.all()

    def get_filter_method(self) -> str:
        """
        Returns preferred Django queryset filter method, e.g. ``istartswith``,
        ``icontains``.

        This will be used to filter querysets doing global searches and single
        column filtering.

        Returns:
            The filter method string.
        """
        return self.FILTER_ISTARTSWITH

    def initialize(self, *args, **kwargs):
        """
        Determine which version of DataTables is being used - there are
        differences in parameters sent by DataTables < 1.10.x
        """
        if 'iSortingCols' in self._querydict:
            self.pre_camel_case_notation = True

    def get_order_columns(self) -> List[str]:
        """
        Return list of columns used for ordering.

        By default returns :py:attr`order_columns` but if these are not defined
        it tries to get columns from the request using the ``columns[i][name]``
        attribute. This requires proper client side definition of columns, eg::

            columns: [
                {
                    name: 'username',
                    data: 'username',
                    orderable: true,
                },
                {
                    name: 'email',
                    data: 'email',
                    orderable: false
                }
            ]
        """
        if self.order_columns or self.pre_camel_case_notation:
            return self.order_columns

        # try to build list of order_columns using request data
        order_columns = []
        for column_def in self.columns_data:
            if column_def['name'] or not self.is_data_list:
                # if self.is_data_list is False then we have a column name in
                # the 'data' attribute, otherwise 'data' attribute is an integer
                # with column index
                if column_def['orderable']:
                    if self.is_data_list:
                        order_columns.append(column_def['name'])
                    else:
                        order_columns.append(column_def.get('data'))
                else:
                    order_columns.append('')
            else:
                # fallback to columns
                order_columns = self._columns
                break

        self.order_columns = order_columns
        return order_columns

    def get_columns(self) -> List[str]:
        """
        Returns the list of columns to be returned in the result set.

        By default returns :py:attr:`columns` but if these are not defined it tries to
        get columns from the request using the ``columns[i][data]`` or
        columns[i][name] attribute.  This requires proper client side definition
        of columns, e.g.::

            columns: [
                {
                    data: 'username'
                },
                {
                    data: 'email'
                }
            ]
        """
        if self.columns or self.pre_camel_case_notation:
            return self.columns

        columns = []
        for column_def in self.columns_data:
            if self.is_data_list:
                # if self.is_data_list is True then 'data' atribute is an
                # integer - column index, so we cannot use it as a column name,
                # let's try 'name' attribute instead
                col_name = column_def['name']
            else:
                col_name = column_def['data']

            if col_name:
                columns.append(col_name)
            else:
                return self.columns

        return columns

    @staticmethod
    def _column_value(obj: Any, key: str) -> Any:
        """
        Returns the value from a queryset item.
        """
        if isinstance(obj, dict):
            return obj.get(key, None)

        return getattr(obj, key, None)

    def _render_column(self, row: Any, column: str) -> str:
        """
        Renders a column on a row. column can be given in a module notation eg.
        ``document.invoice.type``
        """
        # try to find rightmost object
        obj = row
        parts = column.split('.')
        for part in parts[:-1]:
            if obj is None:
                break
            obj = getattr(obj, part)

        # try using get_OBJECT_display for choice fields
        if hasattr(obj, f'get_{parts[-1]}_display'):
            value = getattr(obj, f'get_{parts[-1]}_display')()
        else:
            value = self._column_value(obj, parts[-1])

        if value is None:
            value = self.none_string if value is None else value
        if self.escape_values:
            value = escape(value)

        return value

    def render_column(self, row: Any, column: str) -> str:
        """
        Renders a column on a row. column can be given in a module notation eg.
        ``document.invoice.type``
        """
        value = self._render_column(row, column)
        # if value and hasattr(row, 'get_absolute_url'):
        #     return format_html('<a href="{}">{}</a>', row.get_absolute_url(), value)
        return value

    def ordering(self, qs: models.QuerySet) -> models.QuerySet:
        """
        Get parameters from the request and prepare order by clause.
        """

        # Number of columns that are used in sorting
        sorting_cols = 0
        if self.pre_camel_case_notation:
            try:
                sorting_cols = int(self._querydict.get('iSortingCols', 0))
            except ValueError:
                sorting_cols = 0
        else:
            sort_key = f'order[{sorting_cols}][column]'
            while sort_key in self._querydict:
                sorting_cols += 1
                sort_key = f'order[{sorting_cols}][column]'

        order = []
        order_columns = self.get_order_columns()

        for i in range(sorting_cols):
            # sorting column
            sort_dir = 'asc'
            try:
                if self.pre_camel_case_notation:
                    sort_col = int(self._querydict.get(f'iSortCol_{i}'))
                    # sorting order
                    sort_dir = self._querydict.get(f'sSortDir_{i}')
                else:
                    sort_col = int(self._querydict.get(f'order[{i}][column]'))
                    # sorting order
                    sort_dir = self._querydict.get(f'order[{i}][dir]')
            except ValueError:
                sort_col = 0

            sdir = '-' if sort_dir == 'desc' else ''
            sortcol = order_columns[sort_col]
            if not sortcol:
                continue

            if isinstance(sortcol, list):
                for sc in sortcol:
                    order.append('{0}{1}'.format(sdir, sc.replace('.', '__')))
            else:
                order.append('{0}{1}'.format(sdir, sortcol.replace('.', '__')))

        if order:
            return qs.order_by(*order)
        return qs

    def paging(self, qs: models.QuerySet) -> models.QuerySet:
        if self.pre_camel_case_notation:
            limit = min(int(self._querydict.get('iDisplayLength', 10)), self.max_display_length)
            start = int(self._querydict.get('iDisplayStart', 0))
        else:
            limit = min(int(self._querydict.get('length', 10)), self.max_display_length)
            start = int(self._querydict.get('start', 0))

        # if pagination is disabled ("paging": false)
        if limit == -1:
            return qs

        offset = start + limit

        return qs[start:offset]

    def extract_datatables_column_data(self) -> List[Dict[str, str]]:
        """
        Helper method to extract columns data from request as passed by Datatables 1.10+.
        """
        request_dict = self._querydict
        col_data = []
        if not self.pre_camel_case_notation:
            counter = 0
            data_name_key = f'columns[{counter}][name]'
            while data_name_key in request_dict:
                searchable = request_dict.get(f'columns[{counter}][searchable]') == 'true'
                orderable = request_dict.get(f'columns[{counter}][orderable]') == 'true'
                col_data.append(
                    {
                        'name': request_dict.get(data_name_key),
                        'data': request_dict.get(f'columns[{counter}][data]'),
                        'searchable': searchable,
                        'orderable': orderable,
                        'search.value': request_dict.get(f'columns[{counter}][search][value]'),
                        'search.regex': request_dict.get(f'columns[{counter}][search][regex]')
                    }
                )
                counter += 1
                data_name_key = f'columns[{counter}][name]'
        return col_data

    def prepare_results(
        self,
        qs: models.QuerySet
    ) -> Union[List[List[str]], List[Dict[str, Any]]]:
        data = []
        for item in qs:
            if self.is_data_list:
                data.append([self.render_column(item, column) for column in self._columns])
            else:
                data.append({
                    col_data['data']: self.render_column(item, col_data['data'])
                    for col_data in self.columns_data
                })

        return data

    def handle_exception(self, e):
        logger.exception(str(e))
        raise e

    def get_context_data(self, *args, **kwargs):
        try:
            self.initialize(*args, **kwargs)

            # prepare columns data (for DataTables 1.10+)
            self.columns_data = self.extract_datatables_column_data()

            # determine the response type based on the 'data' field passed from
            # JavaScript https://datatables.net/reference/option/columns.data
            # col['data'] can be an integer (return list) or string (return
            # dictionary) we only check for the first column definition here as
            # there is no way to return list and dictionary at once
            self.is_data_list = True
            if self.columns_data:
                self.is_data_list = False
                try:
                    int(self.columns_data[0]['data'])
                    self.is_data_list = True
                except ValueError:
                    pass

            # prepare list of columns to be returned
            self._columns = self.get_columns()

            qs = self.get_initial_queryset()
            total_records = qs.count()
            qs = self.filter_queryset(qs)
            total_display_records = qs.count()
            qs = self.ordering(qs)
            qs = self.paging(qs)

            # prepare output data
            if self.pre_camel_case_notation:
                ret = {
                    'sEcho': int(self._querydict.get('sEcho', 0)),
                    'iTotalRecords': total_records,
                    'iTotalDisplayRecords': total_display_records,
                    'aaData': self.prepare_results(qs)
                }
            else:
                ret = {
                    'draw': int(self._querydict.get('draw', 0)),
                    'recordsTotal': total_records,
                    'recordsFiltered': total_display_records,
                    'data': self.prepare_results(qs)
                }
            return ret
        except Exception as e:  # pylint: disable=broad-except
            return self.handle_exception(e)


class BaseDatatableView(DatatableMixin, JSONResponseView):
    pass


class DatatableAJAXView(BaseDatatableView):
    """
    This is a JSON view that a DataTables.js table can hit for its AJAX queries.
    """

    def columns(self, querydict: Dict[str, Any]) -> Dict[str, Any]:
        """
        Parse the request we got from the DataTables AJAX request
        (``querydict``) from a list of strings to a more useful nested dict.
        We'll use this to more readily figure out what we need to be doing.

        A single column's data in the querystring we got from datatables looks
        like::

            {
                'columns[4][data]': 'employee_number',
                'columns[4][name]': '',
                'columns[4][orderable]': 'true',
                'columns[4][search][regex]': 'false',
                'columns[4][search][value]': '',
                'columns[4][searchable]': 'true'
            }

        Turn all such data into a dict that looks like::

            {
                'employee_number': {
                    'column_number': 4,
                    'data': 'employee_number',
                    'name': '',
                    'orderable': True,
                    'search': {
                        'regex': False,
                        'value': ''
                    },
                    'searchable': True
                }
            }

        Args:
            querydict: dict of all key, value pairs from the AJAX request.

        Returns:
            The dictionary of column definitions keyed by column name
        """
        by_number = {}
        for key, value in querydict.items():
            if value == 'true':
                value = True
            elif value == 'false':
                value = False
            if key.startswith('columns'):
                parts = key.split('[')
                column_number = parts[1][:-1]
                column_attribute = parts[2][:-1]
                if column_number not in by_number:
                    by_number[column_number] = {
                        'column_number': int(column_number)
                    }
                if len(parts) == 4:
                    if column_attribute not in by_number[column_number]:
                        by_number[column_number][column_attribute] = {parts[3][:-1]: value}
                    else:
                        by_number[column_number][column_attribute][parts[3][:-1]] = value
                else:
                    by_number[column_number][column_attribute] = value
        # Now make the real lookup be by column name
        by_name = {}
        for key, value in by_number.items():
            # 'data' is the column name
            by_name[by_number[key]['data']] = value
        return by_name

    @lru_cache(maxsize=4)
    def searchable_columns(self) -> List[str]:
        """
        Return the list of all column names from our DataTable that are marked
        as "searchable".

        Returns:
            List of searchable columns.
        """
        return [
            key for key, value in self.columns(self._querydict).items()
            if value['searchable']
        ]

    def column_specific_searches(self) -> List[Tuple[str, str]]:
        """
        Look through the request data we got from the DataTable AJAX request for
        any single-column searches.  These will look like
        ``column[$number][search][value]`` in the request.

        Returns:
            A list of 2-tuples that look like ``(column name, search string)``
        """
        return [
            (key, value['search']['value'])
            for key, value in self.columns(self._querydict).items()
            if value['search']['value']
        ]

    def single_column_filter(
        self,
        qs: models.QuerySet,
        column: str,
        value: str
    ) -> models.QuerySet:
        """
        Filter our queryset by a single column.  Columns will be searched by
        ``__icontains``.

        This gets executed when someone runs a search on a single DataTables.js
        column::

            data_table.column($number).search($search_string);

        If you want to implement a different search filter for a particular
        column, add a ``search_COLUMN_column(self, qs, column, value)`` method
        that returns a QuerySet to your subclass and that will be called
        instead.

        Example::

            def search_foobar_column(self, qs, column, value):
                if value == 'something':
                    qs = qs.filter(foobar__attribute=True)
                elif value == 'other_thing':
                    qs = qs.filter(foobar__other_attribute=False)
                return qs

        Args:
            qs: the Django QuerySet
            column: the dataTables data attribute for the column we're interested in
            value: the search string

        Returns:
            An appropriately filtered :py:class:`QuerySet`
        """
        attr_name = f'filter_{column}_column'
        if hasattr(self, attr_name):
            qs = getattr(self, attr_name)(qs, column, value)
        elif column in self.searchable_columns():
            kwarg_name = f'{column}__icontains'
            qs = qs.filter(**{kwarg_name: value})
        return qs

    def search_query(self, qs: models.QuerySet, value: str) -> models.Q:
        """
        Return an ORed Q() object that will search the searchable fields for ``value``

        Args:
            qs: the Django QuerySet
            value: the search string

        Returns:
            A properly formatted :py:class:`Q` object
        """
        # FIXME: we doesn't use qs, so why are we accepting it as a parameter
        query = None
        for column in self.searchable_columns():
            attr_name = f'search_{column}_column'
            if hasattr(self, attr_name):
                q = getattr(self, attr_name)(column, value)
            else:
                kwarg_name = f'{column}__icontains'
                q = Q(**{kwarg_name: value})
            query = query | q if query else q
        return query

    def search(self, qs: models.QuerySet, value: str) -> models.QuerySet:
        """
        Filter our queryset across all of our searchable columns for ``value``.

        This gets executed when someone runs a search over the whole
        DataTables.js table::

            data_table.search($search_string);

        This is what happens when the user uses the "Search" input on the
        DataTable.

        Args:
            qs: the Django QuerySet
            value: the search string

        Returns:
            Our ``qs`` filtered by our searchable columns.
        """
        query = self.search_query(qs, value)
        qs = qs.filter(query).distinct()
        return qs

    def filter_queryset(self, qs: models.QuerySet) -> models.QuerySet:
        """
        We're overriding the default filter_queryset(method) here so we can
        implement proper searches on our pseudo-columns "responded" and
        "disabled", and do column specific searches, as well as doing general
        searches across our regular CharField columns.
        """
        column_searches = self.column_specific_searches()
        if column_searches:
            for column, value in column_searches:
                qs = self.single_column_filter(qs, column, value)
        value = self.request.GET.get('search[value]', None)
        if value:
            qs = self.search(qs, value)
        return qs

    def _render_column(self, row: Any, column: str) -> str:
        """
        When defining our DataTable config, we declare DataTable columns that
        access columns of models of foreign keys on our model with ``__``
        notation.  Convert the ``__`` to ``.`` before trying to render the data
        so that things will work correctly.

        Args:
            row: A Django model instance
            column: the name of the column to render

        Return:
            The rendered column as valid HTML.
        """
        column = re.sub('__', '.', column)
        return super()._render_column(row, column)

    def render_column(self, row: Any, column: str) -> str:
        """
        Return the data to add to the DataTable for column ``column`` of the
        table row corresponding to the model object ``row``.

        If you want to implement special rendering a particular column (for
        instance to display something about an object that doesn't map directly
        to a model column), add a ``render_COLUMN_column(self, row, column)``
        method that returns a string to your subclass and that will be called
        instead.

        Example::

            def render_foobar_column(self, row: Any, column: str) -> str:
                return 'some data'

        Args:
            row: A Django model instance
            column: the name of the column to render

        Return:
            The rendered column as valid HTML.
        """
        attr_name = f'render_{column}_column'
        if hasattr(self, attr_name):
            return getattr(self, attr_name)(row, column)
        return super().render_column(row, column)



from copy import deepcopy
from dataclasses import dataclass, field
import re
from typing import Any, Dict, Iterable, List, Optional, Type, Union, cast

from django.core.exceptions import ImproperlyConfigured

from .base import Block, Container, Link, OrderedList
from .structure import CollapseWidget
from .icons import TablerMenuIcon


path_validator_re = re.compile(
    r'^/\S+$', re.IGNORECASE)


url_validator_re = re.compile(
    r'^(?:http)s?://'  # http:// or https://
    r'(?:(?:[A-Z0-9](?:[A-Z0-9-]{0,61}[A-Z0-9])?\.)+(?:[A-Z]{2,6}\.?|[A-Z0-9-]{2,}\.?)|'  # domain...
    r'localhost|'  # localhost...
    r'\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3})'  # ...or ip
    r'(?::\d+)?'  # optional port
    r'(?:/?|[/?]\S+)$', re.IGNORECASE)


# ==============================
# Functions
# ==============================

def is_url(text: str) -> bool:
    """
    if ``text`` looks like a URL or path, return ``True``.

    Args:
        text: the string to check

    Returns:
        ``True`` if ``text`` looks like a URL or path, ``False`` otherwise.
    """
    return bool(path_validator_re.search(text) or url_validator_re.search(text))


# ==============================
# Dataclasses
# ==============================

@dataclass
class MenuItem:
    """
    A menu item definition for a :py:class:`Menu`.  :py:attr:`title` is
    required, but the other attributes are not.

    Note:
        We're defining this dataclass so that you don't need to build out the
        entire menu system with the appropriate blocks yourself, but instead can
        nest :py:class:`MenuItem` objects and have us do it for you.

    If no :py:attr:`url` is provided, we will consider this item to be a section
    title in the :py:class:`Menu`.
    """

    #: The text for the item.  If ``url`` is not defined, this will define
    #: a heading within the menu
    text: str
    #: this is either the name of a bootstrap icon, or a :py:class:`Block`
    icon: Optional[Union[str, Block]] = None
    #: The URL for the item.  For Django urls, you will typically do something like
    #: ``reverse('myapp:view')`` or ``reverse_lazy('myapp:view')``
    url: Optional[str] = None
    #: a submenu under this menu item
    items: List["MenuItem"] = field(default_factory=list)
    #: Is this the page we're currently on?
    active: bool = False

    @property
    def is_active(self) -> bool:
        """
        Return ``True`` if we or any item in :py:attr:`items` have :py:attr:`active` equal
        to ``True``.

        Returns:
            Whether this item is active.
        """
        status = self.active
        if not status:
            return any(item.is_active for item in self.items)
        return status

    def set_active(self, text: str) -> bool:
        """
        If ``text`` looks like a URL or path, look at our :py:attr:`url` and set
        :py:attr:`active` to ``True`` if they are equal.

        If ``text`` does not look like a URL or path, look at our :py:attr:`text` and set :py:attr:`active` to ``True``.

        If we find no match, set :py:attr:`active` to `False`, and try doing
        :py:meth:`set_active` on our :py:attr:`items`. Stop looking when we find
        the first item that matches ``text``.

        Args:
            text: the value to which to compare our :py:attr:`text`

        Returns:
            If we actually did set someone's :py:attr:`active` to ``True``, return
            ``True``, otherwise return False
        """
        if is_url(text):
            target = self.url
        else:
            target = self.text
        if target == text:
            self.active = True
            return True
        self.active = False
        for item in self.items:
            if item.set_active(text):
                return True
        return False


@dataclass
class BreadcrumbItem:
    """
    A dataclass that is used to register crumbs with :py:class:`BreadcrumbBlock`.
    """
    #: The title for the breadcrumb
    title: str
    #: The optional URL for the breadcrumb
    url: str = None


# ==============================
# Blocks
# ==============================

class BreadcrumbBlock(Block):
    """
    A block that renders a `Bootstrap Breadcrumb
    <https://getbootstrap.com/docs/5.2/components/breadcrumb/>`_ component.

    Typically how you will use this is to define a base set of breadcrumbs
    (which typically just has your home page in it), and then add additional
    breadcrumbs in your views.

    Example::

        from django.views import View
        from wildewdigets import BreadcrumbItem, BreadcrumbBlock

        class BaseBreadCrumbs(BreadcrumbBlock):

            def __init__(self, *args, **kwargs):
                self.add_breadcrumb('Home', url='/home')

        class MyView(View):

            def get_context_data(self, **kwargs):
                kwargs = super().get_context_data(**kwargs)
                breadcrumbs = BaseBreadCrumbs()
                breadcrumbs.add_breadcrumb('MyView')
                kwargs['breadcrumbs'] = breadcrumbs
                return kwargs

    Keyword Args:
        title_class: Apply this string of CSS classes to each breadcrumb title
    """

    tag: str = 'nav'
    aria_attributes: Dict[str, str] = {'label': 'breadcrumb'}

    #: Apply this string of CSS classes to each breadcrumb title
    title_class: Optional[str] = None
    #: The base set of breadcrumb items
    items: List[BreadcrumbItem] = []

    def __init__(self, *args, title_class: str = None, **kwargs):
        self.title_class = title_class if title_class else self.title_class
        super().__init__(*args, **kwargs)
        #: The list of :py:class:`BreadcrumbItem` objects from which we will
        #: build our breadcrumb HTML
        self.items = deepcopy(self.items)
        #: The ``<ol>`` that holds our breadcrumbs
        self.breadcrumbs = OrderedList(name='breadcrumb')
        self.add_block(self.breadcrumbs)

    def add_breadcrumb(self, title: str, url: str = None) -> None:
        """
        Add a new breadcrumb item to this :py:class:`BreadcrumbBlock`.

        Example:

            >>> breadcrumbs = BreadcrumbBlock()
            >>> breadcrumbs.add_breadcrumb('Home', url='/home'))
            >>> breadcrumbs.add_breadcrumb('Project')

            This will produce this HTML::

                <nav aria-label="breadcrumb">
                  <ol class="breadcrumb">
                    <li class="breadcrumb-item"><a href="/home">Home</a></li>
                    <li class="breadcrumb-item active">Project</li>
                  </ol>
                </nav>

        Args:
            title: the text for the breadcrumb item

        Keyword Args:
            url: the url for the breadcrumb item

        """
        self.items.append(BreadcrumbItem(title=title, url=url))

    def get_context_data(self, *args, **kwargs) -> Dict[str, Any]:
        for item in self.items:
            if self.title_class:
                title = Block(item.title, tag='span', css_class=self.title_class)
            else:
                title = item.title
            if item.url:
                block = Link(title, url=item.url)
            else:
                block = title
            self.breadcrumbs.add_block(block, name='breadcrumb-item')
        # Make the last li be active
        self.breadcrumbs.blocks[-1].add_class('active')
        return super().get_context_data(*args, **kwargs)

    def flatten(self) -> str:
        """
        Return a string of the titles of all the breadcrumbs.
        """
        return ' - '.join([item.title for item in self.items])


class NavigationTogglerButton(Block):
    """
    The ``navbar-toggler`` button for hiding and showing menus in responsive
    `Bootstrap Navbar containers
    <https://getbootstrap.com/docs/5.2/components/navbar/#responsive-behaviors>`_
    ,  styled with some Tabler styling. This is the hamburger icon that causes
    the menus to appear and disappear when we drop below a specific viewport
    size.

    Example::

        >>> toggler = NavigationTogglerButton(target='the_target')

    Keyword Args:
        target: the CSS id of the tag to toggle
        label: the ARIA label of this button
    """

    tag: str = 'button'
    block: str = 'navbar-toggler'
    css_class: str = 'collapsed'
    attributes: Dict[str, str] = {'type': 'button'}
    data_attributes: Dict[str, str] = {'toggle': 'collapse'}
    aria_attributes: Dict[str, str] = {'expanded': 'false'}

    #: The CSS id of the tag that this button toggles
    target: Optional[str] = None
    #: The ARIA label for this button
    label: str = 'Toggle navigation'

    def __init__(self, target: str = None, label: str = None, **kwargs):
        self.target = target if target else self.target
        self.label = label if label else self.label
        if not self.target:
            raise ValueError(
                'No target supplied; define it either as a constructor kwarg or as a class attribute'
            )
        super().__init__(**kwargs)
        self._aria_attributes['label'] = self.label
        self._aria_attributes['controls'] = self.target
        self._data_attributes['target'] = f"#{self.target.lstrip('#')}"
        self.add_block(
            Block(tag='span', name='navbar-toggler-icon')
        )


class Navbar(Block):
    """
    A horizontal `Bootstrap Navbar <https://getbootstrap.com/docs/5.2/components/navbar>`_.

    Examples:

        With a subclass::

            class MyMenu(Menu):
                items = [MenuItem(text='One', url='/one', icon='target)]

            class MyNavbar(Navbar):
                branding = LinkedImage(src='/static/branding.png', alt='My Brand', url='#')
                contents = [MyMenu()]

        With :py:class:`Navbar` constructor arguments::

            >>> branding = LinkedImage(src='/static/branding.png', alt='My Brand', url='#')
            >>> items = [MenuItem(text='One', url='/one'), ... ]
            >>> menu = Menu(*items)
            >>> sidebar = Navbar(menu, branding=branding)

        Adding menus later::

            >>> items2 = [MenuItem(text='Foo', url='/foo'), ... ]
            >>> menu2 = Menu(*items)
            >>> sidebar.add_to_menu_section(menu2)

    Note:

        You can add other things to the menu section than just menus.  For instance,
        a search form::

            class SearchFormBlock(Block):
                ...

            class MyMenu(Menu):
                items = [MenuItem(text='One', url='/one', icon='target)]

            class MyNavbar(Navbar):
                branding = LinkedImage(src='/static/branding.png', alt='My Brand', url='#')
                contents = [SearchFormBlock(), MyMenu()]

    Args:
        *blocks: a list of :py:class:`wildewidget.Block` (or subclass) instances to
            be added to this container

    Keyword Args:
        branding: A block that will be displayed at the top of the container.
            A good choice might be a :py:class:`LinkedImage` or :py:class:`Image`
        logo_width: The width for the logo image.  Any valid value
            for CSS ``width`` will be accepted.
        title: A title.  If ``logo`` is ``None``, this will be used in its
            place. If `logo` is not ``None``, this will be used as the image ``alt``
            attribute.
        contents_id: the CSS id of the menu container.  Typically you won't need
            to change this unless you have namespace collisions with other CSS
            ids on your page.

    """
    tag: str = 'aside'
    block: str = 'navbar'

    #: Valid values for ``navbar-expand-``. From
    #:`Navbar: How It Works <https://getbootstrap.com/docs/5.3/components/navbar/#how-it-works>`_
    VALID_BREAKPOINTS: List[str] = [
        'sm',
        'md',
        'lg',
        'xl',
        'xxl'
    ]
    #: Valid background colors.  Note that these are Tabler colors, not Bootstrap
    #: colors.  We use Tabler colors because Tabler also defines an appropriate
    #: foreground color for each background color.
    VALID_BACKGROUND_COLORS: List[str] = [
        'blue',
        'azure',
        'indigo',
        'purple',
        'pink',
        'red',
        'orange',
        'yellow',
        'lime',
        'green',
        'teal',
        'cyan',
    ]

    #: Set to ``True`` to use a dark background instead of light
    dark: bool = False
    #: If :py:attr:`dark` is ``True``, set the background color. Default: ``#1d273b``
    background_color: Optional[str] = None
    #: the CSS id of the menu container.  Typically you won't need
    #: to change this unless you have namespace collisions with other CSS
    #: ids on your page.
    contents_id: str = 'sidebar-menu'
    #: A block that will be displayed at the top of the container.
    #: A good choice might be a :py:class:`LinkedImage` or :py:class:`Image`
    branding: Optional[Block] = None
    #: A list of blocks to include in our sidebar
    contents: Iterable[Block] = []
    #: The viewport size at which our menu container collapses to be hidden,
    #: requiring the hamburger menu to show
    hide_below_viewport: str = 'lg'
    #: The width of our actual navbar
    container_size: str = 'fluid'

    def __init__(
        self,
        *blocks: Block,
        contents_id: str = None,
        branding: Block = None,
        hide_below_viewport: str = None,
        container_size: str = None,
        dark: bool = None,
        background_color: str = None,
        **kwargs
    ):
        self.contents_id = contents_id if contents_id else self.contents_id
        self.hide_below_viewport = hide_below_viewport if hide_below_viewport else self.hide_below_viewport
        self.container_size = container_size if container_size else self.container_size
        self.dark = dark if dark is not None else self.dark
        self.background_color = background_color if background_color is not None else self.background_color
        if self.hide_below_viewport not in self.VALID_BREAKPOINTS:
            raise ValueError(
                f'"{self.hide_below_viewport}" is not a valid breakpoint size. '
                f'Choose from: {", ".join(self.VALID_BREAKPOINTS)}'
            )
        if self.background_color and self.background_color not in self.VALID_BACKGROUND_COLORS:
            raise ValueError(
                f'"{self.background_color}" is not a known color. '
                f'Choose from: {", ".join(self.VALID_BACKGROUND_COLORS)}'
            )

        super().__init__(**kwargs)
        # Set our hamburger menu breakpoint
        self.add_class(f'navbar-expand-{self.hide_below_viewport}')
        # light vs dark
        if self.dark:
            self.add_class('navbar-dark')
        # background color
        if self.background_color:
            self.add_class(f'bg-{self.background_color}')
            self.add_class(f'bg-{self.background_color}-fg')
        # Set our "role" attribute to make us more accessible
        self._attributes['role'] = 'navigation'
        #: Everything inside our sidebar lives in this inner container
        self.inner: Block = Container(size=self.container_size, css_class='ms-0')
        self.add_block(self.inner)
        #: This is the branding block at the start of the navbar
        self.branding: Block = branding if branding else deepcopy(self.branding)
        self.build_brand()
        # The menu toggler button for small viewports
        self.inner.add_block(NavigationTogglerButton(target=self.contents_id))
        #: This is the container for all menus
        self.menu_container: Block = CollapseWidget(css_id=self.contents_id, css_class='navbar-collapse')
        self.inner.add_block(self.menu_container)
        for block in self.contents:
            self.add_to_menu_section(block)

    def add_blocks(self) -> None:
        """
        Overrides :py:meth:`wildewidgets.widgets.base.Block.add_blocks`.

        We're overriding :py:meth:`wildewidgets.widgets.base.Block.add_blocks`
        here so that it does nothing because we need to add menus not to our
        top level tag content, but only to the menu section of ourselves,
        :py;attr:`menu_container`.

        :py:attr:`menu_container` doesn't exist yet when we run the constructor
        for :py:class:`wildewidgets.widgets.base.Block`, which is where
        ``add_blocks`` runs.
        """
        pass

    def build_brand(self) -> None:
        """
        Build our ``.navbar-brand`` block and add it to :py:attr:`inner`, our
        content container.  We're doing this in a method here so we can override
        it in subclasses.
        """
        if self.branding:
            # Add the .navbar-brand class to our branding block to make it work
            # properly within the .navbar
            self.branding.add_class('navbar-brand')

    def add_to_menu_section(self, block: Block) -> None:
        """
        Add a block to the menu specfic section of the sidebar,
        :py:attr:`menu_container`.  The menu specific section will be hidden and
        controlled by the hamburger menu when we drop below a specific viewport.

        Warning:
            Use this to add your blocks to this :py:class:`Navbar` instead of
            using :py:meth:`add_block`.

        Args:
            block: the block to add to the menu section
        """
        self.menu_container.add_block(block)

    def activate(self, text: str) -> bool:
        """
        Loop through all the :py:class:`Menu` blocks in
        :py:attr:`menu_container` and set the first menu item we find that
        matches ``text`` to be active.

        Args:
            text: the value to which to compare to menu items

        Returns:
            If we actually did set an item to be active, return
            ``True``, otherwise return ``False``
        """
        for block in self.menu_container.blocks:
            if isinstance(block, Menu):
                if block.activate(text):
                    return True
        return False


class TablerVerticalNavbar(Navbar):
    """
    Make this a Tabler vertical dark navbar.

    * Make the navbar vertical instead of horizontal
    * :py:attr:`dark` is always ``True``
    * Fix width of sidebar at 15rem, unless :py:attr:`wide` is ``True``, when we set it to 18rem.
    * CSS ``position`` is always ``fixed``.
    * Supply some open/close animations.

    Some caveats:

    * :py:attr:`container_size` is ignored

    Example:

        >>> branding = LinkedImage(
            src='/static/branding.png',
            alt='My Brand',
            url='https://example.com',
            width='100%`
        )
        >>> items = [MenuItem(text='One', url='/one'), ... ]
        >>> menu = TablerMenu(*items)
        >>> sidebar = TablerVerticalNavbar(menu, branding=branding)

    Keyword Args:
        wide: If ``True``, make the navbar width be 18rem instead of 15rem
    """
    block: str = 'navbar navbar-vertical'
    dark: bool = True

    #: Make the navbar 18rem wide instead of 15rem
    wide: bool = False

    def __init__(self, *args, wide: bool = None, **kwargs):
        self.wide = wide if wide is not None else self.wide
        super().__init__(*args, **kwargs)
        if self.wide:
            self._css_class += " navbar-wide"

    def build_brand(self) -> None:
        """
        Overrides :py:meth:`Navbar.build_brand`.

        Wrap our branding in an ``<h1>`` and give that ``<h1>`` the
        ``navbar-brand`` class because that's how Tabler does it.
        """
        brand_container = Block(self.branding, tag='h1', css_class='navbar-brand navbar-brand-autodark')
        self.inner.add_block(brand_container)


class MenuHeading(Block):
    """
    A Tabler menu specific heading.  This is a heading within a :py:class:`Menu`
    that sepearates the actual menu items into groups.

    Typically, you won't use this directly, but instead it will be created for
    you from a :py:class:`MenuItem` specification when :py:attr:`MenuItem.url`
    is ``None`` and :py:attr:`MenuItem.items` is empty.

    Example:

        >>> heading = MenuHeading(text='My Heading')

    Kewyword Args:
        text: the text of the heading
    """

    block: str = 'nav-link nav-subtitle'
    css_class: str = 'nav-link my-1 fw-bold text-uppercase'

    #: The text of the heading
    text: Optional[str] = None

    def __init__(self, text: str = None, **kwargs):
        self.text = text if text else self.text
        super().__init__(**kwargs)
        self.add_block(self.text)


class NavItem(Block):
    """
    An entry in a :py:class:`Menu`, the top level menu class used in a
    :py:class:`Navbar` in the menu section.  This is a label with a link to a
    URL, with an optional icon preceding the label.

    Typically, you won't use this directly, but instead it will be created for
    you from a :py:class:`MenuItem` specification.

    If :py:attr:`url` and the ``url`` kwarg are ``None``, this will be a heading
    instead of an item that links to a URL.

    Examples:

        With keyword arguments::

            >>> icon = TablerMenuIcon(icon='target')
            >>> item = NavItem(text='Page', url='/page', icon=icon)

        With a :py:class:`MenuItem`::

            >>> menu_item = MenuItem(text='Page', url='/page', icon='target')
            >>> item2 = NavItem(item=menu_item)

        Adding a in-menu heading::

            >>> item3 = NavItem(text='My Heading')

    Kewyword Args:
        icon: Either the name of a Bootstrap icon, or a :py:class:`TablerMenuIcon` object
        text: The text for the item
        url: The URL for the item
        active: ``True`` if this represents the page we're currently on
        item: a :py:class:`MenuItem`

    Raises:
        ValueError: one or more of the settings failed validation
    """

    tag: str = 'li'
    block: str = 'nav-item'

    #: Either the name of a Bootstrap icon, or a :py:class:`TablerMenuIcon`
    #: object
    icon: Optional[Union[str, TablerMenuIcon]] = None
    #: The text for the item.
    text: Optional[str] = None
    #: The URL for the item.
    url: Optional[str] = None

    def __init__(
        self,
        text: str = None,
        icon: Union[str, TablerMenuIcon] = None,
        url: str = None,
        active: bool = False,
        item: MenuItem = None,
        **kwargs
    ):
        self.active: bool = active
        if item and (text or icon or url or active):
            raise ValueError('Specify "item" or ("text", "icon", "url", "active"), but not both')
        if item:
            self.text = item.text
            self.icon = item.icon if item.icon else deepcopy(self.icon)
            self.url = item.url if item.url else self.url
            self.active = item.active
        else:
            self.text = text if text else self.text
            self.icon = icon if icon else self.icon
            self.url = url if url else self.url
            if not self.text:
                raise ValueError('"text" is required as either a class attribute or keyword arg')
        super().__init__(**kwargs)
        if self.active:
            self.add_class('active')
        icon_block: Block = None
        if self.icon:
            if isinstance(self.icon, TablerMenuIcon):
                icon_block = self.icon
            else:
                icon_block = TablerMenuIcon(icon=self.icon)
        contents: Block
        if self.url:
            contents = Link(url=self.url, css_class='nav-link')
            if icon_block:
                contents.add_block(icon_block)
            contents.add_block(self.text)
        else:
            contents = MenuHeading(text=self.text)
        self.add_block(contents)


# Submenus:

class ClickableNavDropdownControl(Block):
    """
    This is a variation on :py:class:`NavDropdownControl` for the cases when the
    text that controls the open and close of the related
    :py:class:`DropdownMenu` is also a link that should be able to be clicked
    separately.

    To implement that, this is a ``<div>`` with two different links in it:

    * A first link that link that links our :py:attr:`text` to :py:attr:`url`,
      optionally with an icon.
    * A second link that is just the up/down arrow for the dropdown control.
      Clicking this is what opens the submenu with CSS id :py:attr:`menu_id`.

    Examples:

        No icon:

        >>> control = ClickableNavDropdownControl(
            'the-menu-id',
            text='My Text',
            url='/destination'
        )

        With an icon:

        >>> control2 = ClickableNavDropdownControl(
            'the-menu-id',
            text='My Text',
            url='/destination',
            icon='arrow-down-square-fill'
        )

        With a :py:class:`TablerMenuIcon`:

        >>> arrow_icon = TablerMenuIcon(icon='arrow-down-square-fill')
        >>> control3 = ClickableNavDropdownControl(
            'the-menu-id',
            text='My Text',
            url='/destination',
            icon=arrow_icon
        )

    Args:
        menu_id: The CSS Id to of the dropdown menu that this controls.

    Keyword Args:
        icon: either the name of a Bootstrap icon, or a :py:class:`TablerMenuIcon`
            class or subclass
        text: The label for the dropdown control
        url: The URL for the item.  For Django urls, you will typically do something like
            ``reverse('myapp:view')`` or ``reverse_lazy('myapp:view')``

    Raises:
        ValueError: no ``text`` was supplied
    """

    block: str = 'nav-item--clickable'

    #: Either the name of a Bootstrap icon, or a :py:class:`TablerMenuIcon`
    #: class or subclass
    icon: Optional[Union[str, TablerMenuIcon]] = None
    #: The actual name of the dropdown
    text: Optional[str] = None
    #: The URL to associated with the control
    url: Optional[str] = None

    def __init__(
        self,
        menu_id: str,
        text: str = None,
        icon: Union[str, TablerMenuIcon] = None,
        url: str = None,
        active: bool = False,
        **kwargs
    ):
        #: If this is ``True``, this control itself is active, but nothing
        #: in the related :py:class:`DropdownMenu` is
        self.active: bool = active
        self.text = text if text else self.text
        self.icon = icon if icon else deepcopy(self.icon)
        self.url = url if url else self.url
        if not self.url:
            raise ValueError('"url" is required as either a class attribute of a keyword arg')
        if not self.text:
            raise ValueError('"text" is required as either a class attribute of a keyword arg')
        super().__init__(**kwargs)
        # These classes make the link + control look right
        for klass in ['d-flex', 'flex-row', 'justify-content-between', 'align-items-center']:
            self.add_class(klass)
        self.link = Link(url=self.url, name='nav-link')
        # make the clickable link
        if self.icon:
            self.link.add_block(TablerMenuIcon(icon=self.icon))
        self.link.add_block(self.text)
        # make the actual dropdown control
        self.control = Link(
            css_class='nav-link dropdown-toggle',
            role='button',
            data_attributes={
                'toggle': 'dropdown-ww',  # This data-bs-toggle is targeted by our own javascript
                'target': f'#{menu_id}'
            },
            aria_attributes={
                'expanded': 'false'
            }
        )
        self.add_block(self.link)
        self.add_block(self.control)
        if self.active:
            self.add_class('active')

    def expand(self) -> None:
        """
        Set our ``aria-expanded`` attribute to ``true``.
        """
        self.control._aria_attributes['expanded'] = 'true'

    def collapse(self) -> None:
        """
        Set our ``aria-expanded`` attribute to ``false``.
        """
        self.control._aria_attributes['expanded'] = 'false'


class NavLinkToggle(Link):
    """
    This is a control for a
    :py:class:`wildewidgets.widgets.structure.CollapseWidget` that is some text
    with an open/closed indicator.  Clicking on the text opens or closes the
    ``CollapseWidget``.

    This differs from :py:class:`wildewidgets.widgets.buttons.CollapseButton` in that
    it doesn't have button styling, and has the open/closed indicator.

    Keyword Args:
        icon: Either the name of a Bootstrap icon, or a :py:class:`TablerMenuIcon` object
        text: The text for the item
        active: ``True`` if this represents the page we're currently on
        collapse_id: the CSS id of the :py:class:`wildewidgets.widgets.structure.CollapseWidget`
    """

    block: str = 'nav-link'
    name: str = 'nav-link-toggle'
    data_attributes: Dict[str, str] = {
        'toggle': 'collapse',
    }
    aria_attributes: Dict[str, str] = {
        'expanded': 'false'
    }

    #: Either the name of a Bootstrap icon, or a :py:class:`TablerMenuIcon`
    #: class or subclass
    icon: Optional[Union[str, TablerMenuIcon]] = None
    #: The title for the link
    text: Optional[str] = None
    #: The CSS id of the Collapse that we control
    collapse_id: Optional[str] = None

    def __init__(
        self,
        text: str = None,
        icon: Union[str, TablerMenuIcon] = None,
        active: bool = False,
        collapse_id: str = None,
        **kwargs
    ):
        #: This item is active, but nothing in the related :py:class:`DropdownMenu` is
        self.active: bool = active
        self.text = text if text else self.text
        self.icon = icon if icon else deepcopy(self.icon)
        self.collapse_id = collapse_id if collapse_id else self.collapse_id
        if self.collapse_id is None:
            raise self.RequiredAttrOrKwarg('collapse_id')
        if not self.text:
            raise self.RequiredAttrOrKwarg('text')
        super().__init__(role='button', **kwargs)
        self._data_attributes['target'] = self.collapse_id
        # these classes cause the up/down arrow to be nicely separated from the text
        self.add_class('d-flex')
        self.add_class('flex-row')
        self.add_class('justify-content-between')
        if self.icon:
            if isinstance(self.icon, TablerMenuIcon):
                icon_block = self.icon
            else:
                icon_block = TablerMenuIcon(icon=self.icon)
            self.add_block(icon_block)
        self.add_block(self.text)
        if self.active:
            self.add_class('active')

    def expand(self) -> None:
        """
        Set our ``aria-expanded`` attribute to ``true``.
        """
        self._aria_attributes['expanded'] = 'true'

    def collapse(self) -> None:
        """
        Set our ``aria-expanded`` attribute to ``false``.
        """
        self._aria_attributes['expanded'] = 'false'


class NavDropdownControl(Link):
    """
    This is a dropdown control that opens a submenu from within either a
    :py:class:`Menu` or a :py:class`DropdownMenu` within a :py:class:`Menu`.  It
    consists of a label with optional icon in front, plus a arrow that toggles
    direction when the menu opens and closes.

    Examples:

        No icon:

        >>> control = NavDropdownControl(text='My Control')

        With an icon:

        >>> control2 = NavDropdownControl(text='My Control', icon='arrow-down-square-fill')

        With a :py:class:`TablerMenuIcon`:

        >>> arrow_icon = TablerMenuIcon(icon='arrow-down-square-fill')
        >>> control3 = NavDropdownControl(text='My Control', icon=arrow_icon)

    Keyword Args:
        icon: either the name of a Bootstrap icon, or a :py:class:`TablerMenuIcon`
            class or subclass
        text: The label for the dropdown control
        button_id: The CSS Id to assign to this button.  We need this to tie the
            button to the actual :py:class:`DropdownItem`

    Raises:
        ValueError: no ``text`` was supplied
    """

    block: str = 'nav-link'
    name: str = 'dropdown-toggle'
    data_attributes: Dict[str, str] = {
        'toggle': 'dropdown',
        'auto-close': 'true'
    }
    aria_attributes: Dict[str, str] = {
        'expanded': 'false'
    }
    #: Either the name of a Bootstrap icon, or a :py:class:`TablerMenuIcon`
    #: class or subclass
    icon: Optional[Union[str, TablerMenuIcon]] = None
    #: The actual name of the dropdown
    text: Optional[str] = None
    #: The CSS Id to assign to this button.  We need this to tie the button
    #: to the actual :py:class:`TablerDropdownItem`
    button_id: Optional[str] = None

    def __init__(
        self,
        text: str = None,
        icon: Union[str, TablerMenuIcon] = None,
        active: bool = False,
        button_id: str = None,
        **kwargs
    ):
        #: This item is active, but nothing in the related :py:class:`DropdownMenu` is
        self.active: bool = active
        self.text = text if text else self.text
        self.icon = icon if icon else deepcopy(self.icon)
        self.button_id = button_id if button_id else self.button_id
        if not self.text:
            raise self.RequiredAttrOrKwarg('text')
        if not self.button_id:
            raise self.RequiredAttrOrKwarg('button_id')
        super().__init__(css_id=self.button_id, role='button', **kwargs)
        if self.icon:
            if isinstance(self.icon, TablerMenuIcon):
                icon_block = self.icon
            else:
                icon_block = TablerMenuIcon(icon=self.icon)
            self.add_block(icon_block)
        self.add_block(self.text)
        if self.active:
            self.add_class('active')

    def expand(self) -> None:
        """
        Set our ``aria-expanded`` attribute to ``true``.
        """
        self._aria_attributes['expanded'] = 'true'

    def collapse(self) -> None:
        """
        Set our ``aria-expanded`` attribute to ``false``.
        """
        self._aria_attributes['expanded'] = 'false'


class DropdownItem(Link):
    """
    An entry in a :py:class:`DropdownMenu`, a submenu class used in a
    :py:class:`NavDropdownItem` of a :py:class:`Menu`.  This is a
    label with a link to a URL, with an optional icon preceding the label.

    Typically, you won't use this directly, but instead it will be created for
    you from a :py:class:`MenuItem` specification when :py:attr:`MenuItem.url`
    is not ``None``.

    Examples:

        >>> icon = TablerMenuIcon(icon='target')
        >>> item = DropdownItem(text='Page', url='/page', icon=icon)

        >>> menu_item = MenuItem(text='Page', url='/page', icon='target')
        >>> item = DropdownItem(item=menu_item)

    Kewyword Args:
        icon: Either the name of a Bootstrap icon, or a :py:class:`TablerMenuIcon` object
        text: The text for the item
        active: ``True`` if this represents the page we're currently on
        item: a :py:class:`MenuItem`

    Raises:
        ValueError: one or more of the arguments failed validation
    """
    block: str = 'dropdown-item'

    #: this is either the name of a bootstrap icon, or a :py:class:`TablerMenuIcon`
    #: class or subclass
    icon: Optional[Union[str, TablerMenuIcon]] = None
    #: The text for the item.
    text: Optional[str] = None

    def __init__(
        self,
        text: str = None,
        icon: str = None,
        active: bool = False,
        item: MenuItem = None,
        **kwargs
    ):
        # Does this item represent the page we're on?
        self.active: bool = active
        if item and (text or icon or kwargs.get('url', None)):
            raise ValueError('Specify "item" or ("text", "icon", "url"), but not both')
        if item:
            self.text = item.text
            self.icon = item.icon if item.icon else self.icon
            if item.url:
                kwargs['url'] = item.url
            self.active = item.active
        else:
            self.text = text if text else self.text
            self.icon = icon if icon else self.icon
        if not self.text:
            raise ValueError('"text" is required as either a class attribute or keyword arg')
        if not self.url:
            raise ValueError('"url" is required as either a class attribute or keyword arg')
        super().__init__(**kwargs)
        if self.active:
            self.add_class('active')
        icon_block: Block = None
        if self.icon:
            icon_block = TablerMenuIcon(icon=self.icon, css_class='text-white')
        if icon_block:
            self.add_block(icon_block)
        self.add_block(self.text)


# TODO: tabler dropdown divider

class DropdownMenu(Block):
    """
    A `Tabler dropdown menu <https://preview.tabler.io/docs/dropdowns.html>`_.

    Typically, you won't use this directly, but instead it will be created for
    you from a :py:class:`MenuItem` specification when :py:attr:`MenuItem.url`
    is not ``None``.

    Examples:

        >>> items = [
            MenuItem(text='One', url='/one', icon='1-circle'),
            MenuItem(text='Two', url='/two', icon='2-circle'),
            MenuItem(text='Three', url='/three', icon='3-circle'),
        ]
        >>> button = NavDropdownControl(
            text='My Dropdown Menu',
            icon='arrow-down-square-fill',
            button_id='my-button'
        )
        >>> menu = DropdownMenu(*items, button_id=button.button_id)

    Args:
        *items: the list of :py:class:`MenuItem` objects to insert into
            this menu

    Keyword Args:
        button_id: it CSS id for the button

    Raises:
        ValueError: one or more of the settings failed validation
    """

    block: str = 'dropdown-menu'

    #: A list of :py:class:`MenuItem` objects to add to this dropdown menu
    items: Iterable[MenuItem] = []
    #: The id of the dropdown-toggle button that controls this menu
    button_id: Optional[str] = None

    def __init__(
        self,
        *items: MenuItem,
        button_id: str = None,
        **kwargs
    ):
        self.button_id = button_id if button_id else self.button_id
        if items:
            self.items: Iterable[MenuItem] = items
        else:
            self.items = deepcopy(self.items)
        super().__init__(**kwargs)
        if not self.button_id and not self._css_id:
            raise ValueError('Either "button_id" or "css_id" are required as either class attributes of keyword args')
        if self.button_id:
            self._aria_attributes['labelledby'] = self.button_id
        for item in items:
            self.add_item(item=item)

    def add_item(
        self,
        text: str = None,
        url: str = None,
        icon: Union[str, TablerMenuIcon] = None,
        active: bool = False,
        item: MenuItem = None
    ) -> None:
        """
        Add a new item to this :py:class:`DropdownMenu`.

        Kewyword Args:
            icon: Either the name of a Bootstrap icon, or a :py:class:`TablerMenuIcon` object
            text: The text for the item
            url: The URL for the item
            active: ``True`` if this represents the page we're currently on
            item: a :py:class:`MenuItem`

        Raises:
            ValueError: one or more of the settings failed validation
        """
        if item and (text or icon or url or active):
            raise ValueError('Specify "item" or ("text", "icon", "url"), but not both')
        if not item:
            if text:
                raise ValueError('"text" is required if "item" is not provided')
            item = MenuItem(text=cast(str, text), url=url, icon=icon, active=active)
        if item.items:
            self.add_block(NavDropdownItem(
                *item.items,
                text=item.text,
                url=item.url,
                icon=item.icon
            ))
        else:
            self.add_block(DropdownItem(item=item))

    def show(self) -> None:
        """
        Force this dropdown menu to be shown.
        """
        self.add_class('show')

    def hide(self) -> None:
        """
        Force this dropdown menu to be hidden.
        """
        self.remove_class('show')


class NavDropdownItem(Block):
    """
    An item in a :py:class:`Menu` that opens a dropdown submenu.

    Typically, you won't use this directly, but instead it will be created for
    you from a :py:class:`MenuItem` specification when :py:attr:`MenuItem.items`
    is not empty.

    Examples:

        >>> sub_items = [
            MenuItem(text='One', url='/one', icon='1-circle'),
            MenuItem(text='Two', url='/two', icon='2-circle'),
            MenuItem(text='Three', url='/three', icon='3-circle'),
        ]
        >>> item = NavDropdownItem(*sub_items, text='My Submenu', icon='target')
        >>> menu = Menu(item)


    Args:
        *items: the list of :py:class:`MenuItem` objects to insert into
            this menu

    Keyword Args:
        text: the text for the menu item
        icon: this is either the name of a Bootstrap icon, or a
            :py:class:`TablerMenuIcon` class or subclass

    Raises:
        ValueError: one or more of the settings failed validation
    """

    tag: str = 'li'
    block: str = 'nav-item dropdown'

    #: this is either the name of a bootstrap icon, or a :py:class:`TablerMenuIcon`
    #: class or subclass
    icon: Optional[Union[str, TablerMenuIcon]] = None
    #: The actual name of the dropdown
    text: Optional[str] = None
    #: The URL for the control text.  Use this if you want :py:attr:`text` to be
    #: clickable separately from opening the dropdown menu
    url: Optional[str] = None
    #: The list of items in this dropdown menu
    items: Iterable[MenuItem] = []

    def __init__(
        self,
        *items: MenuItem,
        text: str = None,
        icon: str = None,
        url: str = None,
        active: bool = False,
        **kwargs
    ):

        #: The control for opening the dropdown menu is active, but nothing in the
        #: related :py:class:`DropdownMenu` is active
        self.active = active
        #: The control that opens and closes :py:attr:`menu`
        self.control: Union[NavDropdownControl, ClickableNavDropdownControl]
        #: The :py:class:`DropdownMenu`  that :py:attr:`control` opens
        self.menu: DropdownMenu
        self.text = text if text else self.text
        self.url = url if url else self.url
        self.icon = icon if icon else deepcopy(self.icon)
        if not self.text:
            raise ValueError('"text" is required as either a class attribute of a keyword arg')
        if items:
            self.items: Iterable[MenuItem] = items
        else:
            self.items = deepcopy(self.items)
        super().__init__(**kwargs)
        if self.url:
            # We need to be able to click on the control to get to a page
            menu_id: str = f'nav-menu-{self.text.lower()}'
            menu_id = re.sub('[ ._]', '-', menu_id)
            self.menu = DropdownMenu(*self.items, css_id=menu_id)
            self.control  = (
                ClickableNavDropdownControl(
                    menu_id=menu_id,
                    text=self.text,
                    icon=self.icon,
                    url=self.url,
                    active=self.active
                )
            )
            # FIXME: should we set up aria-labelledby on the menu to point to the control?
        else:
            # We don't need to be able to click on the control to get to a page
            button_id: str = f'nav-item-{self.text.lower()}'
            button_id = re.sub('[ ._]', '-', button_id)
            self.control = NavDropdownControl(
                button_id=button_id,
                text=self.text,
                icon=self.icon,
                active=self.active
            )
            self.menu = DropdownMenu(*self.items, button_id=button_id)
        self.add_block(self.control)
        self.add_block(self.menu)

    def show(self) -> None:
        """
        Show our :py:attr:`menu`.
        """
        self.control.expand()
        self.menu.show()

    def hide(self) -> None:
        """
        Hide our :py:attr:`menu`.
        """
        self.control.collapse()
        self.menu.hide()

    def get_context_data(self, **kwargs) -> Dict[str, Any]:
        """
        Overrides :py:meth:`wildewidgets.widgets.base.Block.get_context_data`.

        Show our menu if either :py:attr:`active` is ``True`` or any item in our
        :py:attr:`menu` is active.  We do this here instead of in our constructor
        so that any menu item activation that happens after instance construction
        time is accounted for.
        """
        is_active: bool = self.active or any(item.is_active for item in self.items)
        if is_active:
            self.show()
        else:
            self.hide()
        return super().get_context_data(**kwargs)


class Menu(Block):
    """
    A ``<div>`` with an optional title and a  `Bootstrap ``ul.navbar-nav``
    <https://getbootstrap.com/docs/5.2/components/navbar/>`_ for use in a
    :py:class:`Navbar`.

    Use this in any of these ways:

    * Subclass :py:class:`Menu` and define :py:attr:`items`
    * Pass in the ``items`` kwargs to the constructor
    * Add items with :py:meth:`add_item`
    * Subclass :py:class:`Menu`, define :py:attr:`items`, and add additional items with
      :py:meth:`add_item`

    Examples:

        Subclassing::

            class MyMenu(Menu):

                items = [
                    MenuItem(text='One', url='/one', icon='1-circle'),
                    MenuItem(text='Two', url='/two', icon='2-circle'),
                    MenuItem(text='Three', url='/three', icon='3-circle'),
                ]

        Constructor::

            >>> menu = Menu(
                MenuItem(text='One', url='/one', icon='1-circle'),
                MenuItem(text='Two', url='/two', icon='2-circle'),
                MenuItem(text='Three', url='/three', icon='3-circle'),
            )

        ``Menu.add_item``::

            >>> menu = Menu()
            >>> menu.add_item(MenuItem(text='One', url='/one', icon='1-circle'))
            >>> menu.add_item(MenuItem(text='Two', url='/two', icon='2-circle'))
            >>> menu.add_item(MenuItem(text='Three', url='/three', icon='3-circle'))

    Args:
        *items: the list of :py:class:`MenuItem` objects to insert into
            this menu

    Keyword Args:
        title: the title for this menu
        title_tag: the HTML tag to use for the title
        title_css_classes: CSS classes to apply to the title

    Raises:
        ValueError: one or more of the settings failed validation
    """

    tag: str = 'div'
    block: str = 'menu'
    css_class: str = 'me-1'

    #: The list of items in this menu
    items: Iterable[MenuItem] = []
    #: The title for this menu
    title: Optional[str] = None
    #: The HTML tag for this title
    title_tag: str = 'h4'
    #: CSS classes to apply to the title
    title_css_classes: str = ''

    def __init__(
        self,
        *items: MenuItem,
        title: str = None,
        title_tag: str = None,
        title_css_classes: str = None,
        **kwargs
    ):
        self.title = title if title else self.title
        self.title_tag = title_tag if title_tag else self.title_tag
        self.title_css_classes = title_css_classes if title_css_classes else self.title_css_classes
        if items:
            self._items: List[MenuItem] = list(items)
        else:
            self._items = list(deepcopy(self.items))
        super().__init__(**kwargs)

    def get_content(self, **kwargs) -> str:
        """
        Actually build out the menu from our list of :py:class:`MenuItems`.  We do
        this in :py:meth:`get_content` instead of in ``__init__`` so that we catch
        any menu items added via :py:meth:`add_item` after instantiation.
        """
        self.build_menu(self._items)
        return super().get_content(**kwargs)

    def build_menu(self, items: Iterable[MenuItem]) -> None:
        """
        Recurse through ``items`` and build out our menu and any submenus.

        For each ``item`` in ``items``, if ``item.items`` is not empty, add a
        submenu, otherwise add simple item.

        Args:
            items: the list of menu items to add to the list
        """
        ul = Block(tag='ul', css_class='navbar-nav')
        if self.title:
            self.add_block(Block(self.title, tag=self.title_tag, css_class=self.title_css_classes + ' menu-title'))
        self.add_block(ul)
        for item in items:
            if item.items:
                ul.add_block(NavDropdownItem(
                    *item.items,
                    text=item.text,
                    icon=item.icon,
                    url=item.url,
                    active=item.active
                ))
            else:
                ul.add_block(NavItem(item=item))

    def add_item(self, item: MenuItem) -> None:
        """
        Add a single :py:class:`MenuItem` to ourselves.

        Args:
            item: the menu item to add to ourselves.
        """
        self._items.append(item)

    def activate(self, text: str) -> bool:
        """
        If ``text`` is not a URL or path, Set :py:attr:`MenuItem.active` to
        ``True`` for the first item we find in our :py:attr:`items` (searching
        recursively) whose :py:attr:`MenuItem.text` matches ``text``.

        If ``text`` is a URL or path, Set :py:attr:`MenuItem.active` to
        ``True`` for the first item we find in our :py:attr:`items` (searching
        recursively) whose :py:attr:`MenuItem.url` matches ``text``.

        Args:
            text: the text to search for among our :py:attr:`items`

        Returns:
            If we found a match, return ``True``, otherwise return ``False``.
        """
        for item in self._items:
            if item.set_active(text):
                return True
        return False


# ==============================
# View mixins
# ==============================

class NavbarMixin:
    """
    This is a Django class based view mixin that offers methods and attributes
    for managing how your :py:class:`Navbar` classes get rendered in your template.

    To use :py:class:`MenuMixin`, you must, in the template for your view,
    render the ``menu`` and ``submenu`` context parameters with the
    ``wildewidgets`` template tag.  These will be the :py:class:`Navbar`
    subclasses that :py:meth:`MenuMixin.get_context_data` adds to the template
    context.   For example::

        <html>
          <head>
            {# wildewdigets css and js requirements #}
          </head>
          <body>
            {% wildewidgets menu %}
            <div class="page">
                ...
                {% wildewidgets submenu %}
                ...
            </div>
          </body>
        </html

    Example:

        This is the most basic example of how to use :py:class:`NavbarMixin` to
        manage a :py:class:`Navbar` with a :py:class:`Menu` inside it and have
        one of the menu items be marked as active::

            from django.views import View
            from django.urls import reverse_lazy
            from wildewidgets import (
                TablerVerticalNavbar,
                Menu,
                MenuItem,
                NavbarMixin
            )

            class MainMenu(Menu):
                items = [
                    MenuItem(
                        text='Projects',
                        icon='bookshelf',
                        url=reverse_lazy('myapp:project--list')
                    ),
                    MenuItem(
                        text='Users',
                        icon='users',
                        url=reverse_lazy('myapp:users--list')
                    )
                ]

            class NavigationSidebar(TablerVerticalNavbar):
                branding = LinkedImage(
                    image_src=static("myapp/images/logo.png"),
                    image_width='100%',
                    image_alt="My App",
                    url="/"
                )

                def __init__(self, *args, **kwargs):
                    super().__init__(*args, **kwargs)
                    self.add_to_menu_section(MainMenu())


            class MyView(NavbarMixin, View):

                navbar_class = NavigationSidebar
                menu_item = 'Projects'

                ...

        This defines our navbar ``NavigationSidebar``, adds ``MainMenu` to it,
        assigns it as our main menu (via ``MyView.navbar_class``), and activates
        the "Projects" item in the main menu (via ``MyView.menu_item``).

        If we wanted to have a secondary navbar also, just define a new
        :py:class:`Navbar` subclass, with a ``secondary_menu_item``::

            class SecondaryMenu(Menu):
                ...

            class ContextualNavbar(Navbar):
                ...

            class MyView(NavbarMixin, View):

                navbar_class = NavigationSidebar
                menu_item = 'Projects'
                secondary_navbar_class = ContextualNavbar
                secondary_menu_item = 'The item'

                ...

    """

    #: The :py:class:`Navbar`` subclass that acts as our primary navigation
    navbar_class: Type[Navbar] = Navbar
    #: The text or URL of an item in one of our :py:attr:`navbar_class` menus to
    #: set active
    menu_item: Optional[str] = None
    #: The :py:class:`Navbar` subclass that holds our secondary menus
    secondary_navbar_class: Optional[Type[Navbar]] = None
    #: The text or URL of an item in one of our
    #: :py:attr:`secondary_navbar_class` menus to set active
    secondary_menu_item: Optional[str] = None

    def get_navbar_class(self) -> Type[Navbar]:
        """
        Return the :py:class:`Navbar` subclass for the main menu.  Overriding
        this will allow you to programmatically determine your at view time.

        Returns:
            The class of the :py:class:`Navbar` subclass to use for our
            main menu.
        """
        return self.navbar_class

    def get_menu_item(self) -> Optional[str]:
        """
        Return the text or URL of an item to set active among the menus
        in our main :py:class:`Navbar` class.

        Returns:
            The class of the :py:class:`Navbar` subclass to use for our
            main menu.
        """
        return self.menu_item

    def get_navbar(self) -> Navbar:
        """
        Instantiate and return our :py:class:`Navbar` subclass for our
        main navbar.

        Returns:
            The instantiated :py:class:`Navbar` subclass instance to use for our
            main navbar.
        """
        navbar_class = self.get_navbar_class()
        if not navbar_class:
            raise ImproperlyConfigured('"navbar_class" must not be None')
        return navbar_class()

    def get_secondary_navbar_class(self) -> Optional[Type[Navbar]]:
        """
        Return our :py:class:`Navbar` subclass for the secondary navbar.

        Returns:
            The class of the :py:class:`Navbar` subclass to use for our
            secondary navbar, if any.
        """
        return self.secondary_navbar_class

    def get_secondary_menu_item(self) -> Optional[str]:
        """
        Return the text of an item to set active in our secondary menu
        """
        return self.secondary_menu_item

    def get_secondary_navbar(self) -> Optional[Navbar]:
        """
        Instantiate and return our :py:class:`Navbar` subclass for our
        secondary navbar.

        Returns:
            The instantiated :py:class:`Navbar` subclass instance to use for our
            secondary menu, if any.
        """
        secondary_navbar_class = self.get_secondary_navbar_class()
        if secondary_navbar_class:
            # pylint is saying that secondary_navbar_class is not callable, which is not true
            return secondary_navbar_class()  # pylint: disable=not-callable
        return None

    def get_context_data(self, **kwargs) -> Dict[str, Any]:
        """
        Add our main navbar and our secondary navbar (if we have one) to the context
        data for our view as the ``menu`` and ``submenu`` kwargs, respectively

        Returns:
            The updated context data.
        """
        kwargs['menu'] = self.get_navbar()
        if menu_item := self.get_menu_item():
            kwargs['menu'].activate(menu_item)
        secondary_navbar = self.get_secondary_navbar()
        if secondary_navbar:
            kwargs['submenu'] = secondary_navbar
            if submenu_item := self.get_secondary_menu_item():
                secondary_navbar.activate(submenu_item)
        return super().get_context_data(**kwargs)  # type: ignore



from typing import List, Optional, Type, TYPE_CHECKING

from django.core.exceptions import ImproperlyConfigured
from django.http import HttpResponseRedirect
from django.views.generic import View
from django.views.generic.base import TemplateView

if TYPE_CHECKING:
    from ..widgets import BaseDataTable


class TableActionFormView(View):
    """
    A view that handles a form action on a table. You just need to define the :py:meth:`process_form_action` method.
    """

    http_method_names: List[str] = ['post']
    url: Optional[str] = None

    def process_form_action(self, action: str, items: List[str]) -> None:
        method_name = f'process_{action}_action'
        if hasattr(self, method_name):
            getattr(self, method_name)(items)

    def post(self, request, *args, **kwargs):
        checkboxes = request.POST.getlist('checkbox')
        action = request.POST.get('action')
        self.process_form_action(action, checkboxes)
        return HttpResponseRedirect(self.url)


class TableView(TemplateView):

    table_class: "Optional[Type[BaseDataTable]]" = None

    def get_context_data(self, **kwargs):
        if not self.table_class:
            raise ImproperlyConfigured(
                f"You must set a table_class attribute on {self.__class__.__name__}"
            )
        kwargs['table'] = self.table_class()  # pylint: disable=not-callable
        return super().get_context_data(**kwargs)



from typing import List, Tuple


class DataTableColumn:

    def __init__(
        self,
        field: str,
        verbose_name: str = None,
        searchable: bool = False,
        sortable: bool = False,
        align: str = 'left',
        head_align: str = 'left',
        visible: bool = True,
        wrap: bool = True
    ):
        self.field = field
        self.verbose_name = verbose_name if verbose_name else self.field.capitalize()
        self.searchable = searchable
        self.sortable = sortable
        self.align = align
        self.head_align = head_align
        self.visible = visible
        self.wrap = wrap


class DataTableFilter:

    def __init__(self, header=None):
        self.header = header
        self.choices: List[Tuple[str, str]] = [('Any', '')]

    def add_choice(self, label: str, value: str) -> None:
        self.choices.append((label, value))


class DataTableStyler:

    def __init__(self, is_row, test_cell, cell_value, css_class, target_cell=None):
        self.is_row = is_row
        self.test_cell = test_cell
        self.cell_value = cell_value
        self.css_class = css_class
        self.target_cell = target_cell
        self.test_index = 0
        self.target_index = 0


class DataTableForm:

    def __init__(self, table):
        if table.has_form_actions():
            self.is_visible: bool = True
        else:
            self.is_visible = False
        self.actions = table.get_form_actions()
        self.url = table.form_url



#!/usr/bin/env python
# -*- coding: utf-8 -*-
from copy import copy
from dataclasses import dataclass
from typing import List, Optional, Dict, Any, Union

from .base import Widget, Block, WidgetStream
from .buttons import LinkButton, FormButton
from .headers import PageHeader, WidgetListLayoutHeader


@dataclass
class WidgetIndexItem:
    widget: Widget
    title: str
    icon: str = 'gear'


class WidgetIndex(Block):
    template_name: str = 'wildewidgets/widget_index.html'
    block: str = "widget-index"
    entries: List[WidgetIndexItem] = []

    def __init__(self, *args, **kwargs):
        entries = kwargs.pop('entries', copy(self.entries))
        super().__init__(*args, **kwargs)
        self._entries = entries

    @property
    def is_empty(self) -> bool:
        return len(self._entries) == 0

    def add_widget(self, widget, title=None, icon=None):
        """
        Add ``widget`` the index with title ``title`` and icon ``icon``.

        If ``title`` is ``None``, look for a title on ``widget.title``.  If that is also ``None``,
        default to the name of the widget class.

        If ``icon`` is ``None``, look for an icon on ``widget.icon``.  If that is also ``None``,
        default to the Bootstrap Icons "gear" icon.
        """
        if hasattr(widget, 'is_visible') and not widget.is_visible:
            return
        item = WidgetIndexItem(
            widget=widget,
            title=getattr(widget, 'title', widget.__class__.__name__),
            icon=getattr(widget, 'icon', 'gear')
        )
        if title is not None:
            item.title = title
        if icon is not None:
            item.icon = icon
        self._entries.append(item)

    def get_context_data(self, *args, **kwargs):
        context = super().get_context_data(*args, **kwargs)
        context['entries'] = self._entries
        return context


class WidgetListSidebarWidget(Block):

    template_name: str = 'wildewidgets/widget-list--sidebar.html'

    block: str = "widget-list__sidebar"
    css_class: Optional[str] = None
    actions: List[Widget] = []
    bare_widgets: List[Widget] = []

    class Widgets(WidgetStream):
        block: str = 'widget-list__sidebar__widgets'

    class Actions(WidgetStream):
        css_class: str = 'px-3 py-4 d-flex flex-column align-items-stretch border bg-white shadow-sm'

    def __init__(
        self,
        *args,
        title: str = None,
        width: int = 3,
        breakpoint: str = 'xl',
        bare_widgets: List[Widget] = None,
        actions: List[Widget] = None,
        **kwargs
    ):
        if title is not None:
            self.title = title
        if self.css_class is None:
            self.css_class = ''
        if breakpoint:
            self.css_class += f"col-{breakpoint}-{width}"
        else:
            self.css_class += f"col-{width}"
        super().__init__(*args, **kwargs)
        actions = actions if actions is not None else self.actions
        bare_widgets = bare_widgets if bare_widgets is not None else self.bare_widgets
        self.widget_index = WidgetIndex()
        self._widgets = WidgetListSidebarWidget.Widgets(widgets=bare_widgets)
        self._actions = WidgetListSidebarWidget.Actions(widgets=actions)
        self._actions.block = f"{self.block}__actions"

    def add_link_button(self, text: str, url: str, **kwargs):
        """
        Add :class:`wildewidgets.LinkButton` to the sidebar.

        :param text: use this as the button text
        :type text: str
        :param url: the URL for the link button
        :type url: str

        You may also use any of the keyword arguments for :class:`wildewidgets.LinkButton`.
        """
        kwargs['text'] = text
        kwargs['url'] = url
        if 'css_class' in kwargs:
            kwargs['css_class'] = f'{kwargs["css_class"]} w-100'
        else:
            kwargs['css_class'] = 'w-100'
        self.add_actions_widget(LinkButton(**kwargs))

    def add_form_button(self, text: str, action: str, **kwargs):
        """
        Add :class:`wildewidgets.FormButton` to the sidebar.

        :param text: use this as the button text
        :type text: str
        :param url: the URL for the link button
        :type url: str

        You may also use any of the keyword arguments for :class:`wildewidgets.FormButton`.
        """
        kwargs['text'] = text
        kwargs['action'] = action
        if 'css_class' in kwargs:
            kwargs['css_class'] = f'{kwargs["css_class"]} w-100'
        else:
            kwargs['css_class'] = 'w-100'
        if 'button_css_class' in kwargs:
            kwargs['button_css_class'] = f'{kwargs["button_css_class"]} w-100'
        else:
            kwargs['button_css_class'] = 'w-100'
        self.add_actions_widget(FormButton(**kwargs))

    def add_widget(self, widget: Widget):
        """
        Add a widget to the sidebar outside the Actions box.
        """
        self._widgets.add_widget(widget)

    def add_actions_widget(self, widget: Widget):
        """
        Add a widget to the sidebar inside the Actions box.
        """
        self._actions.add_widget(widget)

    def add_widget_to_index(self, widget: Widget, title: Optional[str] = None, icon: Optional[str] = None):
        """
        Add ``widget`` the index with title ``title`` and icon ``icon``.

        If ``title`` is ``None``, look for a title on ``widget.title``.  If that is also ``None``,
        default to the name of the widget class.

        If ``icon`` is ``None``, look for an icon on ``widget.icon``.  If that is also ``None``,
        default to the Bootstrap Icons "gear" icon.
        """
        self.widget_index.add_widget(widget, title=title, icon=icon)

    def get_context_data(self, *args, **kwargs) -> Dict[str, Any]:
        context = super().get_context_data(*args, **kwargs)
        context['title'] = self.title
        if not self._widgets.is_empty:
            context['widgets'] = self._widgets
            self.widget_index._css_class = 'mt-5'
        if not self._actions.is_empty:
            context['actions'] = self._actions
            self.widget_index._css_class = 'mt-5'
        if not self.widget_index.is_empty:
            context['widget_index'] = self.widget_index
        return context


class WidgetListMainWidget(Block):
    template_name: str = 'wildewidgets/widget-list--main.html'
    block: str = "widget-list__main"

    css_class: Optional[str] = 'col'
    entry_css_class: Optional[str] = 'shadow bg-white'
    entry_title_css_class: Optional[str] = 'font-weight-bold'
    entries: List[WidgetIndexItem] = []

    def __init__(self, *args, **kwargs):
        entries = kwargs.pop('entries', copy(self.entries))
        self._entry_css_class = kwargs.pop('entry_css_class', self.entry_css_class)
        self._entry_title_css_class = kwargs.pop('entry_title_css_class', self.entry_title_css_class)
        super().__init__(*args, **kwargs)
        self._entries: List[WidgetIndexItem] = entries

    def add_widget(self, widget: Widget, title: Optional[Union[str, Widget]] = None):
        if title is not None:
            widget.title = title
        widget_title = widget.get_title()
        if not isinstance(widget_title, Widget):
            header = WidgetListLayoutHeader(header_text=widget.title)
            widget.title = header
        self._entries.append(widget)

    def get_context_data(self, *args, **kwargs) -> Dict[str, Any]:
        context = super().get_context_data(*args, **kwargs)
        context['entries'] = self._entries
        context['entry_title_css_class'] = self._entry_title_css_class
        context['entry_css_class'] = self._entry_css_class
        return context


class WidgetListLayout(Block):
    """
    This class provides a two column layout. The first column is the sidebar
    containing links to the various widgets in the right column, and the second
    column is the main content, consisting of the contained widgets vertically
    stacked.

    Example:

        >>> layout = WidgetListLayout('My Page')
        >>> layout.add_sidebar_form_button('Update', reverse('core:thing--update'))
        >>> layout.add_widget(Widget1(), title='another title')
        >>> layout.add_widget(Widget2(), title='the title', icon='stuff')
        >>> layout.add_widget(Widget3())

    Keyword Args:
        title: The title of the widget.
        sidebar_title: the title of the sidebar column
        sidebar_width: the width in columns of the sidebar column

    """
    template_name: str = 'wildewidgets/widget-list.html'
    sidebar_title: str = 'Actions'
    # Number of columns
    sidebar_width: int = 3
    sidebar_breakpoint: str = 'xl'

    def __init__(
        self,
        title: str,
        sidebar_title: str = None,
        sidebar_width: int = None,
        sidebar_breakpoint: str = None,
        **kwargs
    ) -> None:
        """
        """
        super().__init__(**kwargs)
        self.title = title
        self.header: Widget = PageHeader(header_text=title)
        self.sidebar: Widget = WidgetListSidebarWidget(
            title=sidebar_title if sidebar_title is not None else self.sidebar_title,
            width=sidebar_width if sidebar_width is not None else self.sidebar_width,
            breakpoint=sidebar_breakpoint if sidebar_breakpoint is not None else self.sidebar_breakpoint,
        )
        self.main = WidgetListMainWidget()
        self.modals: List[Widget] = []

    def add_widget(self, widget: Widget, title: Optional[str] = None, icon: Optional[str] = None) -> None:
        self.sidebar.add_widget_to_index(widget, title=title, icon=icon)
        self.main.add_widget(widget, title=title)

    def add_modal(self, modal: Widget) -> None:
        self.modals.append(modal)

    def add_sidebar_link_button(self, text: str, url: str, **kwargs) -> None:
        self.sidebar.add_link_button(text, url, **kwargs)

    def add_sidebar_form_button(self, text: str, action: str, **kwargs) -> None:
        self.sidebar.add_form_button(text, action, **kwargs)

    def add_sidebar_widget(self, widget: Widget) -> None:
        self.sidebar.add_actions_widget(widget)

    def add_sidebar_bare_widget(self, widget: Widget) -> None:
        self.sidebar.add_widget(widget)

    def get_context_data(self, *args, **kwargs) -> Dict[str, Any]:
        context = super().get_context_data(*args, **kwargs)
        context['title'] = self.title
        context['sidebar'] = self.sidebar
        context['sidebar_width'] = self.sidebar_width
        context['main'] = self.main
        context['modals'] = self.modals
        context['header'] = self.header
        return context
